import os
import tempfile

import pytest
from numpy.testing import assert_equal

from brian2 import numpy as np
from brian2.spatialneuron import *
from brian2.tests.utils import assert_allclose
from brian2.units import DimensionMismatchError, cm, second, um
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_attributes_soma():
    soma = Soma(diameter=10 * um)
    assert isinstance(soma, Morphology)
    # Single compartment
    assert soma.n == 1
    assert soma.total_sections == 1
    assert soma.total_compartments == 1
    with pytest.raises(TypeError):
        len(soma)  # ambiguous
    # Compartment attributes
    assert_equal(soma.diameter, [10] * um)
    assert_equal(soma.length, [10] * um)
    assert_equal(soma.distance, [0] * um)
    assert_equal(soma.end_distance, 0 * um)
    assert soma.r_length_1 > 1 * cm
    assert soma.r_length_2 > 1 * cm
    assert_equal(soma.area, np.pi * soma.diameter**2)
    assert_allclose(soma.volume, 1.0 / 6.0 * np.pi * (10 * um) ** 3)

    # No coordinates were specified
    assert soma.start_x is None
    assert soma.start_y is None
    assert soma.start_z is None
    assert soma.x is None
    assert soma.y is None
    assert soma.z is None
    assert soma.end_x is None
    assert soma.end_y is None
    assert soma.end_z is None


@pytest.mark.codegen_independent
def test_attributes_soma_coordinates():
    # Specify only one of the coordinates
    xyz = {"x", "y", "z"}
    for coord in xyz:
        kwds = {coord: 5 * um}
        soma = Soma(diameter=10 * um, **kwds)
        # Length shouldn't change (not defined by coordinates but by the diameter)
        assert_equal(soma.length, [10] * um)
        assert_equal(soma.distance, [0] * um)

        # Coordinates should be specified now, with 0 values for the other
        # coordinates
        for other_coord in xyz - {coord}:
            assert_equal(getattr(soma, f"start_{other_coord}"), [0] * um)
            assert_equal(getattr(soma, other_coord), [0] * um)
            assert_equal(getattr(soma, f"end_{other_coord}"), [0] * um)

        assert_equal(getattr(soma, f"start_{coord}"), [5] * um)
        assert_equal(getattr(soma, coord), [5] * um)
        assert_equal(getattr(soma, f"end_{coord}"), [5] * um)

    # Specify all coordinates
    soma = Soma(diameter=10 * um, x=1 * um, y=2 * um, z=3 * um)
    # Length shouldn't change (not defined by coordinates but by the diameter)
    assert_equal(soma.length, [10] * um)
    assert_equal(soma.distance, [0] * um)

    assert_equal(soma.start_x, 1 * um)
    assert_equal(soma.x, 1 * um)
    assert_equal(soma.end_x, 1 * um)
    assert_equal(soma.start_y, 2 * um)
    assert_equal(soma.y, 2 * um)
    assert_equal(soma.end_y, 2 * um)
    assert_equal(soma.start_z, 3 * um)
    assert_equal(soma.z, 3 * um)
    assert_equal(soma.end_z, 3 * um)


@pytest.mark.codegen_independent
def test_attributes_cylinder():
    n = 10
    cylinder = Cylinder(n=n, diameter=10 * um, length=200 * um)
    assert isinstance(cylinder, Morphology)
    # Single section with 10 compartments
    assert cylinder.n == n
    assert cylinder.total_sections == 1
    assert cylinder.total_compartments == n
    with pytest.raises(TypeError):
        len(cylinder)  # ambiguous

    # Compartment attributes
    assert_equal(cylinder.diameter, np.ones(n) * 10 * um)
    assert_equal(cylinder.length, np.ones(n) * 20 * um)
    assert_equal(cylinder.distance, np.arange(n) * 20 * um + 10 * um)
    assert_equal(cylinder.end_distance, 200 * um)
    # TODO: r_length
    assert_allclose(cylinder.area, np.pi * cylinder.diameter * cylinder.length)
    assert_allclose(
        cylinder.volume, 1.0 / 4.0 * np.pi * cylinder.diameter**2 * cylinder.length
    )

    # No coordinates were specified
    assert cylinder.start_x is None
    assert cylinder.start_y is None
    assert cylinder.start_z is None
    assert cylinder.x is None
    assert cylinder.y is None
    assert cylinder.z is None
    assert cylinder.end_x is None
    assert cylinder.end_y is None
    assert cylinder.end_z is None


@pytest.mark.codegen_independent
def test_attributes_cylinder_coordinates():
    # Specify only the end-point of the section
    n = 10
    # Specify only one of the coordinates
    xyz = {"x", "y", "z"}
    for coord in xyz:
        kwds = {coord: [0, 200] * um}
        cylinder = Cylinder(n=n, diameter=10 * um, **kwds)
        assert_equal(cylinder.diameter, np.ones(n) * 10 * um)
        assert_equal(cylinder.length, np.ones(n) * 20 * um)
        assert_equal(cylinder.distance, np.arange(n) * 20 * um + 10 * um)
        assert_equal(cylinder.end_distance, 200 * um)

        # Coordinates should be specified now, with 0 values for the other
        # coordinates
        for other_coord in xyz - {coord}:
            assert_equal(getattr(cylinder, f"start_{other_coord}"), np.zeros(n) * um)
            assert_equal(getattr(cylinder, other_coord), np.zeros(n) * um)
            assert_equal(getattr(cylinder, f"end_{other_coord}"), np.zeros(n) * um)

        assert_equal(getattr(cylinder, f"start_{coord}"), np.arange(n) * 20 * um)
        assert_equal(getattr(cylinder, coord), np.arange(n) * 20 * um + 10 * um)
        assert_equal(
            getattr(cylinder, f"end_{coord}"), np.arange(n) * 20 * um + 20 * um
        )

    # Specify all coordinates
    val = [0, 200.0 / np.sqrt(3.0)] * um
    cylinder = Cylinder(n=n, diameter=10 * um, x=val, y=val, z=val)

    assert_equal(cylinder.diameter, np.ones(n) * 10 * um)
    assert_allclose(cylinder.length, np.ones(n) * 20 * um)
    assert_allclose(cylinder.distance, np.arange(n) * 20 * um + 10 * um)
    assert_allclose(cylinder.end_distance, 200 * um)

    for coord in ["x", "y", "z"]:
        assert_allclose(getattr(cylinder, f"start_{coord}"), np.arange(n) * val[1] / n)
        assert_allclose(
            getattr(cylinder, coord), np.arange(n) * val[1] / n + 0.5 * val[1] / n
        )
        assert_allclose(
            getattr(cylinder, f"end_{coord}"), np.arange(n) * val[1] / n + val[1] / n
        )


@pytest.mark.codegen_independent
def test_attributes_section():
    n = 10
    # No difference to a cylinder
    sec = Section(n=n, diameter=np.ones(n + 1) * 10 * um, length=np.ones(n) * 20 * um)
    cyl = Cylinder(n=1, diameter=10 * um, length=0 * um)  # dummy cylinder
    cyl.child = sec
    assert isinstance(sec, Morphology)
    # Single section with 10 compartments
    assert sec.n == n
    assert sec.total_sections == 1
    assert sec.total_compartments == n
    with pytest.raises(TypeError):
        len(sec)  # ambiguous

    # Compartment attributes
    assert_allclose(sec.diameter, np.ones(n) * 10 * um)
    assert_allclose(sec.length, np.ones(n) * 20 * um)
    assert_allclose(sec.distance, np.arange(n) * 20 * um + 10 * um)
    assert_allclose(sec.end_distance, 200 * um)
    # TODO: r_length
    assert_allclose(
        sec.area, np.pi * 0.5 * (sec.start_diameter + sec.end_diameter) * sec.length
    )
    assert_allclose(sec.volume, 1.0 / 4.0 * np.pi * sec.diameter**2 * sec.length)

    # No coordinates were specified
    assert sec.start_x is None
    assert sec.start_y is None
    assert sec.start_z is None
    assert sec.x is None
    assert sec.y is None
    assert sec.z is None
    assert sec.end_x is None
    assert sec.end_y is None
    assert sec.end_z is None


@pytest.mark.codegen_independent
def test_attributes_section_coordinates_single():
    # Specify only the end-point of the section  (no difference to cylinder)
    n = 10
    # Specify only one of the coordinates
    xyz = {"x", "y", "z"}
    for coord in xyz:
        kwds = {coord: np.linspace(0 * um, 200 * um, n + 1)}
        sec = Section(n=n, diameter=np.ones(n + 1) * 10 * um, **kwds)
        cyl = Cylinder(n=1, diameter=10 * um, length=0 * um)  # dummy cylinder
        cyl.child = sec
        assert_equal(sec.diameter, np.ones(n) * 10 * um)
        assert_equal(sec.length, np.ones(n) * 20 * um)
        assert_equal(sec.distance, np.arange(n) * 20 * um + 10 * um)
        assert_equal(sec.end_distance, 200 * um)

        # Coordinates should be specified now, with 0 values for the other
        # coordinates
        for other_coord in xyz - {coord}:
            assert_equal(getattr(sec, f"start_{other_coord}"), np.zeros(n) * um)
            assert_equal(getattr(sec, other_coord), np.zeros(n) * um)
            assert_equal(getattr(sec, f"end_{other_coord}"), np.zeros(n) * um)

        assert_equal(getattr(sec, f"start_{coord}"), np.arange(n) * 20 * um)
        assert_equal(getattr(sec, coord), np.arange(n) * 20 * um + 10 * um)
        assert_equal(getattr(sec, f"end_{coord}"), np.arange(n) * 20 * um + 20 * um)

    # Specify all coordinates
    val = 200.0 / np.sqrt(3.0) * um
    sec = Section(
        n=n,
        diameter=np.ones(n + 1) * 10 * um,
        x=np.linspace(0 * um, val, n + 1),
        y=np.linspace(0 * um, val, n + 1),
        z=np.linspace(0 * um, val, n + 1),
    )
    cyl = Cylinder(n=1, diameter=10 * um, length=0 * um)
    cyl.child = sec
    assert_equal(sec.diameter, np.ones(n) * 10 * um)
    assert_allclose(sec.length, np.ones(n) * 20 * um)
    assert_allclose(sec.distance, np.arange(n) * 20 * um + 10 * um)
    assert_allclose(sec.end_distance, 200 * um)

    for coord in ["x", "y", "z"]:
        assert_allclose(getattr(sec, f"start_{coord}"), np.arange(n) * val / n)
        assert_allclose(getattr(sec, coord), np.arange(n) * val / n + 0.5 * val / n)
        assert_allclose(getattr(sec, f"end_{coord}"), np.arange(n) * val / n + val / n)


@pytest.mark.codegen_independent
def test_attributes_section_coordinates_all():
    n = 3
    # Specify all coordinates
    sec = Section(
        n=n,
        diameter=[10, 10, 10, 10] * um,
        x=[10, 11, 11, 11] * um,
        y=[100, 100, 101, 101] * um,
        z=[1000, 1000, 1000, 1001] * um,
    )

    assert_equal(sec.diameter, np.ones(n) * 10 * um)
    assert_allclose(sec.length, np.ones(n) * um)
    assert_allclose(sec.distance, np.arange(n) * um + 0.5 * um)
    assert_allclose(sec.end_distance, 3 * um)

    assert_allclose(sec.start_x, [10, 11, 11] * um)
    assert_allclose(sec.x, [10.5, 11, 11] * um)
    assert_allclose(sec.end_x, [11, 11, 11] * um)
    assert_allclose(sec.start_y, [100, 100, 101] * um)
    assert_allclose(sec.y, [100, 100.5, 101] * um)
    assert_allclose(sec.end_y, [100, 101, 101] * um)
    assert_allclose(sec.start_z, [1000, 1000, 1000] * um)
    assert_allclose(sec.z, [1000, 1000, 1000.5] * um)
    assert_allclose(sec.end_z, [1000, 1000, 1001] * um)

    # Specify varying diameters
    sec = Section(
        n=n,
        diameter=[20, 10, 5, 2.5] * um,
        x=[0, 1, 1, 1] * um,
        y=[0, 0, 1, 1] * um,
        z=[0, 0, 0, 1] * um,
    )
    assert_allclose(sec.start_diameter, [20, 10, 5] * um)
    # diameter at midpoint
    assert_allclose(sec.diameter, 0.5 * (sec.start_diameter + sec.end_diameter))
    assert_allclose(sec.end_diameter, [10, 5, 2.5] * um)
    # TODO: Check area and volume


def _check_tree_cables(morphology, coordinates=False):
    # number of compartments per section
    assert morphology.n == 10
    assert morphology["1"].n == 5
    assert morphology["2"].n == 5
    assert morphology["21"].n == 5
    assert morphology["22"].n == 5
    # number of compartments per subtree
    assert morphology.total_compartments == 30
    assert morphology["1"].total_compartments == 5
    assert morphology["2"].total_compartments == 15
    assert morphology["21"].total_compartments == 5
    assert morphology["22"].total_compartments == 5
    # number of sections per subtree
    assert morphology.total_sections == 5
    assert morphology["1"].total_sections == 1
    assert morphology["2"].total_sections == 3
    assert morphology["21"].total_sections == 1
    assert morphology["22"].total_sections == 1
    # Check that distances (= distance to root at electrical midpoint)
    # correctly follow the tree structure
    assert_allclose(morphology.distance, np.arange(10) * 10 * um + 5 * um)
    assert_allclose(
        morphology["2"].distance, 100 * um + np.arange(5) * 10 * um + 5 * um
    )
    assert_allclose(
        morphology["21"].distance, 150 * um + np.arange(5) * 10 * um + 5 * um
    )
    assert_allclose(morphology.end_distance, 100 * um)
    assert_allclose(morphology["1"].end_distance, 200 * um)
    assert_allclose(morphology["2"].end_distance, 150 * um)
    assert_allclose(morphology["21"].end_distance, 200 * um)
    assert_allclose(morphology["22"].end_distance, 200 * um)
    # Check that section diameters are correctly inherited from the parent
    # sections
    assert_allclose(morphology["1"].start_diameter, [10, 8, 6, 4, 2] * um)
    assert_allclose(morphology["22"].start_diameter, [5, 4, 3, 2, 1] * um)

    if coordinates:
        # Coordinates should be absolute
        # section: cable
        assert_allclose(morphology.start_x, np.arange(10) * 10 * um)
        assert_allclose(morphology.x, np.arange(10) * 10 * um + 5 * um)
        assert_allclose(morphology.end_x, np.arange(10) * 10 * um + 10 * um)
        assert_allclose(morphology.y, np.zeros(10) * um)
        assert_allclose(morphology.z, np.zeros(10) * um)
        # section: cable['1']
        step = 20 / np.sqrt(2) * um
        assert_allclose(morphology["1"].start_x, 100 * um + np.arange(5) * step)
        assert_allclose(morphology["1"].x, 100 * um + np.arange(5) * step + step / 2)
        assert_allclose(morphology["1"].end_x, 100 * um + np.arange(5) * step + step)
        assert_allclose(morphology["1"].start_y, np.arange(5) * step)
        assert_allclose(morphology["1"].y, np.arange(5) * step + step / 2)
        assert_allclose(morphology["1"].end_y, np.arange(5) * step + step)
        assert_allclose(morphology["1"].z, np.zeros(5) * um)
        # section: cable['2']
        step = 10 / np.sqrt(2) * um
        assert_allclose(morphology["2"].start_x, 100 * um + np.arange(5) * step)
        assert_allclose(morphology["2"].x, 100 * um + np.arange(5) * step + step / 2)
        assert_allclose(morphology["2"].end_x, 100 * um + np.arange(5) * step + step)
        assert_allclose(morphology["2"].start_y, -np.arange(5) * step)
        assert_allclose(morphology["2"].y, -(np.arange(5) * step + step / 2))
        assert_allclose(morphology["2"].end_y, -(np.arange(5) * step + step))
        assert_allclose(morphology["2"].z, np.zeros(5) * um)
        # section: cable ['21']
        step = 10 / np.sqrt(2) * um
        assert_allclose(
            morphology["21"].start_x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step,
        )
        assert_allclose(
            morphology["21"].x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step + step / 2,
        )
        assert_allclose(
            morphology["21"].end_x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step + step,
        )
        assert_allclose(morphology["21"].start_y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["21"].y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["21"].end_y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["21"].start_z, np.arange(5) * step)
        assert_allclose(morphology["21"].z, np.arange(5) * step + step / 2)
        assert_allclose(morphology["21"].end_z, np.arange(5) * step + step)
        # section: cable['22']
        step = 10 / np.sqrt(2) * um
        assert_allclose(
            morphology["22"].start_x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step,
        )
        assert_allclose(
            morphology["22"].x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step + step / 2,
        )
        assert_allclose(
            morphology["22"].end_x,
            100 * um + 50 / np.sqrt(2) * um + np.arange(5) * step + step,
        )
        assert_allclose(morphology["22"].start_y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["22"].y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["22"].end_y, -np.ones(5) * 50 / np.sqrt(2) * um)
        assert_allclose(morphology["22"].start_z, -np.arange(5) * step)
        assert_allclose(morphology["22"].z, -(np.arange(5) * step + step / 2))
        assert_allclose(morphology["22"].end_z, -(np.arange(5) * step + step))


@pytest.mark.codegen_independent
def test_tree_cables_schematic():
    cable = Cylinder(n=10, diameter=10 * um, length=100 * um)
    cable.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    cable.R = Cylinder(n=5, diameter=5 * um, length=50 * um)
    cable.RL = Cylinder(n=5, diameter=2.5 * um, length=50 * um)
    cable.RR = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    _check_tree_cables(cable)


@pytest.mark.codegen_independent
def test_tree_cables_coordinates():
    # The lengths of the sections should be identical to the previous test
    cable = Cylinder(n=10, x=[0, 100] * um, diameter=10 * um)
    cable.L = Section(
        n=5,
        diameter=[10, 8, 6, 4, 2, 0] * um,
        x=np.linspace(0, 100, 6) / np.sqrt(2) * um,
        y=np.linspace(0, 100, 6) / np.sqrt(2) * um,
    )
    cable.R = Cylinder(
        n=5, diameter=5 * um, x=[0, 50] * um / np.sqrt(2), y=[0, -50] * um / np.sqrt(2)
    )
    cable.RL = Cylinder(
        n=5, diameter=2.5 * um, x=[0, 50] * um / np.sqrt(2), z=[0, 50] * um / np.sqrt(2)
    )
    cable.RR = Section(
        n=5,
        diameter=[5, 4, 3, 2, 1, 0] * um,
        x=np.linspace(0, 50, 6) * um / np.sqrt(2),
        z=np.linspace(0, -50, 6) * um / np.sqrt(2),
    )

    _check_tree_cables(cable, coordinates=True)


@pytest.mark.codegen_independent
def test_tree_cables_from_points():
    # The coordinates should be identical to the previous test
    # fmt: off
    points = [ # cable
              (1,  None, 0,                  0,                0,              10, -1),
              (2,  None, 10,                 0,                0,              10,  1),
              (3,  None, 20,                 0,                0,              10,  2),
              (4,  None, 30,                 0,                0,              10,  3),
              (5,  None, 40,                 0,                0,              10,  4),
              (6,  None, 50,                 0,                0,              10,  5),
              (7,  None, 60,                 0,                0,              10,  6),
              (8,  None, 70,                 0,                0,              10,  7),
              (9,  None, 80,                 0,                0,              10,  8),
              (10, None, 90,                 0,                0,              10,  9),
              (11, None, 100,                0,                0,              10,  10),
              # cable.L  (using automatic names)
              (12, None, 100+20/np.sqrt(2),  20/np.sqrt(2),    0,              8 ,  11),
              (13, None, 100+40/np.sqrt(2),  40/np.sqrt(2),    0,              6 ,  12),
              (14, None, 100+60/np.sqrt(2),  60/np.sqrt(2),    0,              4 ,  13),
              (15, None, 100+80/np.sqrt(2),  80/np.sqrt(2),    0,              2 ,  14),
              (16, None, 100+100/np.sqrt(2), 100/np.sqrt(2),   0,              0 ,  15),
              # cable.R  (using automatic names)
              (17, None, 100+10/np.sqrt(2),  -10/np.sqrt(2),   0,              5 ,  11),
              (18, None, 100+20/np.sqrt(2),  -20/np.sqrt(2),   0,              5 ,  17),
              (19, None, 100+30/np.sqrt(2),  -30/np.sqrt(2),   0,              5 ,  18),
              (20, None, 100+40/np.sqrt(2),  -40/np.sqrt(2),   0,              5 ,  19),
              (21, None, 100+50/np.sqrt(2),  -50/np.sqrt(2),   0,              5 ,  20),
              # cable.RL (using explicit names)
              (22, 'L' , 100+60/np.sqrt(2),  -50/np.sqrt(2),   10/np.sqrt(2),  2.5, 21),
              (23, 'L' , 100+70/np.sqrt(2),  -50/np.sqrt(2),   20/np.sqrt(2),  2.5, 22),
              (24, 'L' , 100+80/np.sqrt(2),  -50/np.sqrt(2),   30/np.sqrt(2),  2.5, 23),
              (25, 'L' , 100+90/np.sqrt(2),  -50/np.sqrt(2),   40/np.sqrt(2),  2.5, 24),
              (26, 'L' , 100+100/np.sqrt(2),  -50/np.sqrt(2),  50/np.sqrt(2),  2.5, 25),
              # cable.RR (using explicit names)
              (27, 'R' , 100+60/np.sqrt(2),  -50/np.sqrt(2),   -10/np.sqrt(2), 4,   21),
              (28, 'R' , 100+70/np.sqrt(2),  -50/np.sqrt(2),   -20/np.sqrt(2), 3,   27),
              (29, 'R' , 100+80/np.sqrt(2),  -50/np.sqrt(2),   -30/np.sqrt(2), 2,   28),
              (30, 'R' , 100+90/np.sqrt(2),  -50/np.sqrt(2),   -40/np.sqrt(2), 1,   29),
              (31, 'R' , 100+100/np.sqrt(2),  -50/np.sqrt(2),  -50/np.sqrt(2), 0,   30),
    ]
    # fmt: on
    cable = Morphology.from_points(points)

    # Check that the names are used
    assert cable.L.n == 5
    assert cable.R.n == 5
    assert cable.RL.n == 5
    assert cable.RR.n == 5
    _check_tree_cables(cable, coordinates=True)


def test_tree_cables_from_swc():
    swc_content = """
# Test file
1   0  0  0  0  5  -1
2   0  10  0  0  5  1
3   0  20  0  0  5  2
4   0  30  0  0  5  3
5   0  40  0  0  5  4
6   0  50  0  0  5  5
7   0  60  0  0  5  6
8   0  70  0  0  5  7
9   0  80  0  0  5  8
10   0  90  0  0  5  9
11   0  100  0  0  5  10
12   2  114.14213562373095  14.142135623730949  0  4  11
13   2  128.2842712474619  28.284271247461898  0  3  12
14   2  142.42640687119285  42.426406871192846  0  2  13
15   2  156.5685424949238  56.568542494923797  0  1  14
16   2  170.71067811865476  70.710678118654741  0  0  15
17   2  107.07106781186548  -7.0710678118654746  0  2.5  11
18   2  114.14213562373095  -14.142135623730949  0  2.5  17
19   2  121.21320343559643  -21.213203435596423  0  2.5  18
20   2  128.2842712474619  -28.284271247461898  0  2.5  19
21   2  135.35533905932738  -35.35533905932737  0  2.5  20
22   2  142.42640687119285  -35.35533905932737  7.0710678118654746  1.25  21
23   2  149.49747468305833  -35.35533905932737  14.142135623730949  1.25  22
24   2  156.5685424949238  -35.35533905932737  21.213203435596423  1.25  23
25   2  163.63961030678928  -35.35533905932737  28.284271247461898  1.25  24
26   2  170.71067811865476  -35.35533905932737  35.35533905932737  1.25  25
27   2  142.42640687119285  -35.35533905932737  -7.0710678118654746  2  21
28   2  149.49747468305833  -35.35533905932737  -14.142135623730949  1.5  27
29   2  156.5685424949238  -35.35533905932737  -21.213203435596423  1  28
30   2  163.63961030678928  -35.35533905932737  -28.284271247461898  0.5  29
31   2  170.71067811865476  -35.35533905932737  -35.35533905932737  0  30
"""
    tmp_filename = tempfile.mktemp("cable_morphology.swc")
    with open(tmp_filename, "w") as f:
        f.write(swc_content)
    cable = Morphology.from_file(tmp_filename)
    os.remove(tmp_filename)
    _check_tree_cables(cable, coordinates=True)


def _check_tree_soma(morphology, coordinates=False, use_cylinders=True):
    # number of compartments per section
    assert morphology.n == 1
    assert morphology["1"].n == 5
    assert morphology["2"].n == 5

    # number of compartments per subtree
    assert morphology.total_compartments == 11
    assert morphology["1"].total_compartments == 5
    assert morphology["2"].total_compartments == 5

    # number of sections per subtree
    assert morphology.total_sections == 3
    assert morphology["1"].total_sections == 1
    assert morphology["2"].total_sections == 1

    assert_allclose(morphology.diameter, [30] * um)

    # Check that distances (= distance to root at midpoint)
    # correctly follow the tree structure
    # Note that the soma does add nothing to the distance
    assert_equal(morphology.distance, 0 * um)
    assert_allclose(morphology["1"].distance, np.arange(5) * 20 * um + 10 * um)
    assert_allclose(morphology["2"].distance, np.arange(5) * 10 * um + 5 * um)
    assert_allclose(morphology.end_distance, 0 * um)
    assert_allclose(morphology["1"].end_distance, 100 * um)
    assert_allclose(morphology["2"].end_distance, 50 * um)

    assert_allclose(morphology.diameter, 30 * um)
    assert_allclose(morphology["1"].start_diameter, [8, 8, 6, 4, 2] * um)
    assert_allclose(morphology["1"].diameter, [8, 7, 5, 3, 1] * um)
    assert_allclose(morphology["1"].end_diameter, [8, 6, 4, 2, 0] * um)
    assert_allclose(morphology["2"].start_diameter, np.ones(5) * 5 * um)
    assert_allclose(morphology["2"].diameter, np.ones(5) * 5 * um)
    assert_allclose(morphology["2"].end_diameter, np.ones(5) * 5 * um)

    if coordinates:
        # Coordinates should be absolute
        # section: soma
        assert_allclose(morphology.start_x, 100 * um)
        assert_allclose(morphology.x, 100 * um)
        assert_allclose(morphology.end_x, 100 * um)
        assert_allclose(morphology.y, 0 * um)
        assert_allclose(morphology.z, 0 * um)
        # section: cable['1']
        step = 20 / np.sqrt(2) * um
        assert_allclose(morphology["1"].start_x, 100 * um + np.arange(5) * step)
        assert_allclose(morphology["1"].x, 100 * um + np.arange(5) * step + step / 2)
        assert_allclose(morphology["1"].end_x, 100 * um + np.arange(5) * step + step)
        assert_allclose(morphology["1"].start_y, np.arange(5) * step)
        assert_allclose(morphology["1"].y, np.arange(5) * step + step / 2)
        assert_allclose(morphology["1"].end_y, np.arange(5) * step + step)
        assert_allclose(morphology["1"].z, np.zeros(5) * um)
        # section: cable['2']
        step = 10 / np.sqrt(2) * um
        assert_allclose(morphology["2"].start_x, 100 * um + np.arange(5) * step)
        if use_cylinders:
            assert_allclose(
                morphology["2"].x, 100 * um + np.arange(5) * step + step / 2
            )
        assert_allclose(morphology["2"].end_x, 100 * um + np.arange(5) * step + step)
        assert_allclose(morphology["2"].start_y, -np.arange(5) * step)
        if use_cylinders:
            assert_allclose(morphology["2"].y, -(np.arange(5) * step + step / 2))
        assert_allclose(morphology["2"].end_y, -(np.arange(5) * step + step))
        if use_cylinders:
            assert_allclose(morphology["2"].z, np.zeros(5) * um)


@pytest.mark.codegen_independent
def test_tree_soma_schematic():
    soma = Soma(diameter=30 * um)
    soma.L = Section(
        n=5, diameter=[8, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    soma.R = Cylinder(n=5, diameter=5 * um, length=50 * um)

    _check_tree_soma(soma)


@pytest.mark.codegen_independent
def test_tree_soma_coordinates():
    soma = Soma(diameter=30 * um, x=100 * um)
    soma.L = Section(
        n=5,
        diameter=[8, 8, 6, 4, 2, 0] * um,
        x=np.linspace(0, 100, 6) / np.sqrt(2) * um,
        y=np.linspace(0, 100, 6) / np.sqrt(2) * um,
    )  # tapering truncated cones
    soma.R = Cylinder(
        n=5, diameter=5 * um, x=[0, 50] * um / np.sqrt(2), y=[0, -50] * um / np.sqrt(2)
    )

    _check_tree_soma(soma, coordinates=True)


@pytest.mark.codegen_independent
def test_tree_soma_from_points():
    # The coordinates should be identical to the previous test
    # fmt: on
    points = [  # soma
        (1, "soma", 100, 0, 0, 30, -1),
        # soma.L
        (2, "L", 100 + 20 / np.sqrt(2), 20 / np.sqrt(2), 0, 8, 1),
        (3, "L", 100 + 40 / np.sqrt(2), 40 / np.sqrt(2), 0, 6, 2),
        (4, "L", 100 + 60 / np.sqrt(2), 60 / np.sqrt(2), 0, 4, 3),
        (5, "L", 100 + 80 / np.sqrt(2), 80 / np.sqrt(2), 0, 2, 4),
        (6, "L", 100 + 100 / np.sqrt(2), 100 / np.sqrt(2), 0, 0, 5),
        # soma.R
        (7, "R", 100 + 10 / np.sqrt(2), -10 / np.sqrt(2), 0, 5, 1),
        (8, "R", 100 + 20 / np.sqrt(2), -20 / np.sqrt(2), 0, 5, 7),
        (9, "R", 100 + 30 / np.sqrt(2), -30 / np.sqrt(2), 0, 5, 8),
        (10, "R", 100 + 40 / np.sqrt(2), -40 / np.sqrt(2), 0, 5, 9),
        (11, "R", 100 + 50 / np.sqrt(2), -50 / np.sqrt(2), 0, 5, 10),
    ]
    # fmt: on
    cable = Morphology.from_points(points)
    _check_tree_soma(cable, coordinates=True, use_cylinders=False)


@pytest.mark.codegen_independent
def test_tree_soma_from_points_3_point_soma():
    # The coordinates should be identical to the previous test
    # fmt: off
    points = [ # soma
              (1,  'soma', 100,                0,                0,              30, -1),
              (2,  'soma', 100,               15,                0,              30,  1),
              (3,  'soma', 100,              -15,                0,              30,  1),
              # soma.L
              (4,  'L'   , 100+20/np.sqrt(2),  20/np.sqrt(2),    0,              8 ,  1),
              (5,  'L'   , 100+40/np.sqrt(2),  40/np.sqrt(2),    0,              6 ,  4),
              (6,  'L'   , 100+60/np.sqrt(2),  60/np.sqrt(2),    0,              4 ,  5),
              (7,  'L'   , 100+80/np.sqrt(2),  80/np.sqrt(2),    0,              2 ,  6),
              (8,  'L'   , 100+100/np.sqrt(2), 100/np.sqrt(2),   0,              0 ,  7),
              # soma.R
              (9,  'R'   , 100+10/np.sqrt(2),  -10/np.sqrt(2),   0,              5 ,  1),
              (10, 'R'   , 100+20/np.sqrt(2),  -20/np.sqrt(2),   0,              5 ,  9),
              (11, 'R'   , 100+30/np.sqrt(2),  -30/np.sqrt(2),   0,              5 ,  10),
              (12, 'R'   , 100+40/np.sqrt(2),  -40/np.sqrt(2),   0,              5 ,  11),
              (13, 'R'   , 100+50/np.sqrt(2),  -50/np.sqrt(2),   0,              5 ,  12),
    ]
    # fmt: on
    cable = Morphology.from_points(points)
    _check_tree_soma(cable, coordinates=True, use_cylinders=False)
    # The first compartment should be a spherical soma!
    assert isinstance(cable, Soma)


@pytest.mark.codegen_independent
def test_tree_soma_from_points_3_point_soma_incorrect():
    # Inconsistent diameters
    # fmt: off
    points = [ # soma
              (1,  'soma', 100,                0,                0,              30, -1),
              (2,  'soma', 100,               15,                0,              28,  1),
              (3,  'soma', 100,              -15,                0,              30,  1),
              # soma.L
              (4,  'L'   , 100+20/np.sqrt(2),  20/np.sqrt(2),    0,              8 ,  1),
              (5,  'L'   , 100+40/np.sqrt(2),  40/np.sqrt(2),    0,              6 ,  4),
              (6,  'L'   , 100+60/np.sqrt(2),  60/np.sqrt(2),    0,              4 ,  5),
              (7,  'L'   , 100+80/np.sqrt(2),  80/np.sqrt(2),    0,              2 ,  6),
              (8,  'L'   , 100+100/np.sqrt(2), 100/np.sqrt(2),   0,              0 ,  7)
    ]
    # fmt: on
    with pytest.raises(ValueError):
        Morphology.from_points(points)

    # Inconsistent coordinates
    # fmt: off
    points = [  # soma
        (1, 'soma', 100, 0, 0, 30, -1),
        (2, 'soma', 100, 15, 0, 30, 1),
        (3, 'soma', 100, -16, 0, 30, 1),
        # soma.L
        (4, 'L', 100 + 20 / np.sqrt(2), 20 / np.sqrt(2), 0, 8, 1),
        (5, 'L', 100 + 40 / np.sqrt(2), 40 / np.sqrt(2), 0, 6, 4),
        (6, 'L', 100 + 60 / np.sqrt(2), 60 / np.sqrt(2), 0, 4, 5),
        (7, 'L', 100 + 80 / np.sqrt(2), 80 / np.sqrt(2), 0, 2, 6),
        (8, 'L', 100 + 100 / np.sqrt(2), 100 / np.sqrt(2), 0, 0, 7)
    ]
    # fmt: on
    with pytest.raises(ValueError):
        Morphology.from_points(points)


@pytest.mark.codegen_independent
def test_tree_soma_from_points_somas_not_at_start():
    # A single soma in the middle (ununusal but possible)
    points = [  # dendrite
        (1, "dend", 100, 0, 0, 10, -1),
        (2, "dend", 100, 0, 0, 10, 1),
        (3, "dend", 100, 0, 0, 10, 2),
        (4, "soma", 100, 0, 0, 30, 3),
        (5, "dend2", 130, 0, 0, 10, 4),
        (6, "dend2", 160, 0, 0, 10, 5),
    ]
    morpho = Morphology.from_points(points)
    assert morpho.total_sections == 3
    assert morpho.total_compartments == 5

    # Several somata (probably an error)
    points = [  # dendrite
        (1, "dend", 0, 0, 0, 10, -1),
        (2, "dend", 30, 0, 0, 10, 1),
        (3, "dend", 60, 0, 0, 10, 2),
        (4, "soma", 90, 0, 0, 30, 3),
        (5, "dend2", 120, 70, 0, 10, 4),
        (6, "dend2", 150, 70, 0, 10, 5),
        (7, "soma", 180, 40, 0, 30, 6),
    ]
    with catch_logs() as logs:
        morpho = Morphology.from_points(points)
    assert len(logs) == 1
    assert logs[0][1].endswith("soma_compartments")
    assert morpho.total_sections == 4
    assert morpho.total_compartments == 6


@pytest.mark.codegen_independent
def test_tree_soma_from_swc():
    swc_content = """
# Test file
1    1  100  0  0  15  -1
2   2  114.14213562373095  14.142135623730949  0  4  1
3   2  128.2842712474619  28.284271247461898  0  3  2
4   2  142.42640687119285  42.426406871192846  0  2  3
5   2  156.5685424949238  56.568542494923797  0  1  4
6   2  170.71067811865476  70.710678118654741  0  0  5
7   2  107.07106781186548  -7.0710678118654746  0  2.5  1
8   2  114.14213562373095  -14.142135623730949  0  2.5  7
9   2  121.21320343559643  -21.213203435596423  0  2.5  8
10   2  128.2842712474619  -28.284271247461898  0  2.5  9
11   2  135.35533905932738  -35.35533905932737  0  2.5  10
"""
    tmp_filename = tempfile.mktemp("cable_morphology.swc")
    with open(tmp_filename, "w") as f:
        f.write(swc_content)
    soma = Morphology.from_file(tmp_filename)
    os.remove(tmp_filename)
    _check_tree_soma(soma, coordinates=True, use_cylinders=False)


@pytest.mark.codegen_independent
def test_tree_soma_from_swc_3_point_soma():
    swc_content = """
# Test file
1    1  100  0  0  15  -1
2    1  100  15  0  15  1
3    1  100  -15  0  15  1
4   2  114.14213562373095  14.142135623730949  0  4  1
5   2  128.2842712474619  28.284271247461898  0  3  4
6   2  142.42640687119285  42.426406871192846  0  2  5
7   2  156.5685424949238  56.568542494923797  0  1  6
8   2  170.71067811865476  70.710678118654741  0  0  7
9   2  107.07106781186548  -7.0710678118654746  0  2.5  1
10   2  114.14213562373095  -14.142135623730949  0  2.5  9
11   2  121.21320343559643  -21.213203435596423  0  2.5  10
12   2  128.2842712474619  -28.284271247461898  0  2.5  11
13   2  135.35533905932738  -35.35533905932737  0  2.5  12
"""
    tmp_filename = tempfile.mktemp("cable_morphology.swc")
    with open(tmp_filename, "w") as f:
        f.write(swc_content)
    soma = Morphology.from_file(tmp_filename)
    os.remove(tmp_filename)
    _check_tree_soma(soma, coordinates=True, use_cylinders=False)


@pytest.mark.codegen_independent
def test_tree_soma_from_swc_3_point_soma_incorrect_points():
    swc_content = """
# Test file
1    1  100  0  0  15  -1
2    1  100  15  0  15  1
3    1  100  -10  0  15  1
4   2  114.14213562373095  14.142135623730949  0  4  1
5   2  128.2842712474619  28.284271247461898  0  3  4
6   2  142.42640687119285  42.426406871192846  0  2  5
7   2  156.5685424949238  56.568542494923797  0  1  6
8   2  170.71067811865476  70.710678118654741  0  0  7
9   2  107.07106781186548  -7.0710678118654746  0  2.5  1
10   2  114.14213562373095  -14.142135623730949  0  2.5  9
11   2  121.21320343559643  -21.213203435596423  0  2.5  10
12   2  128.2842712474619  -28.284271247461898  0  2.5  11
13   2  135.35533905932738  -35.35533905932737  0  2.5  12
"""
    tmp_filename = tempfile.mktemp("cable_morphology.swc")
    with open(tmp_filename, "w") as f:
        f.write(swc_content)
    with pytest.raises(ValueError, match=r".*radius is 15.000um.*"):
        Morphology.from_file(tmp_filename)
    os.remove(tmp_filename)


@pytest.mark.codegen_independent
def test_tree_soma_from_swc_3_point_soma_incorrect_radius():
    swc_content = """
# Test file
1    1  100  0  0  15  -1
2    1  100  15  0  10  1
3    1  100  -15  0  15  1
4   2  114.14213562373095  14.142135623730949  0  4  1
5   2  128.2842712474619  28.284271247461898  0  3  4
6   2  142.42640687119285  42.426406871192846  0  2  5
7   2  156.5685424949238  56.568542494923797  0  1  6
8   2  170.71067811865476  70.710678118654741  0  0  7
9   2  107.07106781186548  -7.0710678118654746  0  2.5  1
10   2  114.14213562373095  -14.142135623730949  0  2.5  9
11   2  121.21320343559643  -21.213203435596423  0  2.5  10
12   2  128.2842712474619  -28.284271247461898  0  2.5  11
13   2  135.35533905932738  -35.35533905932737  0  2.5  12
"""
    tmp_filename = tempfile.mktemp("cable_morphology.swc")
    with open(tmp_filename, "w") as f:
        f.write(swc_content)
    with pytest.raises(ValueError, match=r".*not all the diameters are identical.*"):
        Morphology.from_file(tmp_filename)
    os.remove(tmp_filename)


@pytest.mark.codegen_independent
def test_construction_incorrect_arguments():
    ### Morphology
    dummy_self = Soma(10 * um)  # To allow testing of Morphology.__init__
    with pytest.raises(TypeError):
        Morphology.__init__(dummy_self, n=1.5)
    with pytest.raises(ValueError):
        Morphology.__init__(dummy_self, n=0)
    with pytest.raises(TypeError):
        Morphology.__init__(dummy_self, "filename.swc")

    ### Soma
    with pytest.raises(DimensionMismatchError):
        Soma(10)
    with pytest.raises(TypeError):
        Soma([10, 20] * um)
    with pytest.raises(TypeError):
        Soma(x=[10, 20] * um)
    with pytest.raises(TypeError):
        Soma(y=[10, 20] * um)
    with pytest.raises(TypeError):
        Soma(z=[10, 20] * um)
    with pytest.raises(DimensionMismatchError):
        Soma(x=10)
    with pytest.raises(DimensionMismatchError):
        Soma(y=10)
    with pytest.raises(DimensionMismatchError):
        Soma(z=10)

    ### Cylinder
    # Diameter can only be single value
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=[10, 20] * um, length=100 * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=[10, 20, 30] * um, length=100 * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=np.ones(3, 2) * um, length=100 * um)
    # Length can only be single value
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, length=[10, 20] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, length=[10, 20, 30] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, length=np.ones(3, 2) * um)
    # Coordinates have to be two values
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, x=[10] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, x=[10, 20, 30] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, y=[10] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, y=[10, 20, 30] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, z=[10] * um)
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, z=[10, 20, 30] * um)
    # Need either coordinates or lengths
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um)
    # But not both
    with pytest.raises(TypeError):
        Cylinder(n=3, diameter=10 * um, length=30 * um, x=[0, 30] * um)

    ### Section
    # Diameter have to be n+1 values
    with pytest.raises(TypeError):
        Section(n=3, diameter=10 * um, length=np.ones(3) * 10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=[10, 20, 30] * um, length=np.ones(3) * 10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4, 2) * um, length=np.ones(3) * 10 * um)
    # Length have to be n values
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, length=10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, length=[10, 20] * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, length=np.ones(3, 2) * um)
    # Coordinates have to be n+1 values
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, x=10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, x=[10, 20, 30] * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, y=10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, y=[10, 20, 30] * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, z=10 * um)
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um, z=[10, 20, 30] * um)
    # Need either coordinates or lengths
    with pytest.raises(TypeError):
        Section(n=3, diameter=np.ones(4) * 10 * um)
    # But not both
    with pytest.raises(TypeError):
        Section(
            n=3,
            diameter=np.ones(4) * 10 * um,
            length=[10, 20, 30] * um,
            x=[0, 10, 20, 30] * um,
        )


@pytest.mark.codegen_independent
def test_from_points_long_chain():
    # in previous versions, this led to a recursion error
    points = [(1, "soma", 0, 0, 0, 1, -1)]
    for i in range(2, 10000):
        points.append((i, "dend", 0 + i * 10, 0, 0, 5, i - 1))
    morph = Morphology.from_points(points)
    assert (
        morph.total_compartments == 10000 - 1
    )  # compartments are in-between the points
    assert morph.total_sections == 2


@pytest.mark.codegen_independent
def test_from_points_minimal():
    points = [(1, "soma", 10, 20, 30, 30, -1)]
    morph = Morphology.from_points(points)
    assert morph.total_compartments == 1
    assert_allclose(morph.diameter, 30 * um)
    assert_allclose(morph.x, 10 * um)
    assert_allclose(morph.y, 20 * um)
    assert_allclose(morph.z, 30 * um)


@pytest.mark.codegen_independent
def test_from_points_incorrect():
    # The coordinates should be identical to the previous test
    # fmt: off
    points = [
              (1,  None, 0,  0, 0, 10, -1),
              (2,  None, 10, 0, 0, 10,  1),
              (2,  None, 20, 0, 0, 10,  2),
    ]
    points2 = [
              (1,  None, 0,  0, 0, 10, -1),
              (2,  None, 10, 0, 0, 10,  1),
              (3,  None, 20, 0, 0, 10,  3),
    ]
    points3 = [
              (1,  None, 0,  0, 0, 10, -1),
              (2,  None, 10, 0, 0, 10,  1),
              (3,  None, 20, 0, 0, 10,  4),
    ]
    points4 = [
              (1,  0,    0,  0, 10, -1),
              (2,  10,   0,  0, 10,  1),
              (3,  20,   0,  0, 10,  2),
    ]
    with pytest.raises(ValueError):
         Morphology.from_points(points)
    with pytest.raises(ValueError):
         Morphology.from_points(points2)
    with pytest.raises(ValueError):
         Morphology.from_points(points3)
    with pytest.raises(ValueError):
         Morphology.from_points(points4)


@pytest.mark.codegen_independent
def test_subtree_deletion():
    soma = Soma(diameter=30 * um)
    first_dendrite = Cylinder(n=5, diameter=5 * um, length=50 * um)
    second_dendrite = Cylinder(n=5, diameter=5 * um, length=50 * um)
    second_dendrite.L = Cylinder(n=5, diameter=5 * um, length=50 * um)
    second_dendrite.R = Cylinder(n=5, diameter=5 * um, length=50 * um)
    soma.dend1 = first_dendrite
    soma.dend2 = second_dendrite
    soma.dend3 = Cylinder(n=5, diameter=5 * um, length=50 * um)
    soma.dend3.L = Cylinder(n=5, diameter=5 * um, length=50 * um)
    soma.dend3.L.L = Cylinder(n=5, diameter=5 * um, length=50 * um)

    assert soma.total_compartments == 36

    del soma.dend1
    assert soma.total_compartments == 31
    with pytest.raises(AttributeError):
        soma.dend1
    with pytest.raises(AttributeError):
        delattr(soma, "dend1")
    with pytest.raises(AttributeError):
        soma.__delitem__("dend1")
    assert first_dendrite not in soma.children

    del soma["dend2"]
    assert soma.total_compartments == 16
    with pytest.raises(AttributeError):
        soma.dend2
    assert second_dendrite not in soma.children

    del soma.dend3.LL
    assert soma.total_compartments == 11
    with pytest.raises(AttributeError):
        soma.dend3.LL
    with pytest.raises(AttributeError):
        soma.dend3.L.L


@pytest.mark.codegen_independent
def test_subgroup_indices():
    morpho = Soma(diameter=30 * um)
    morpho.L = Cylinder(length=10 * um, diameter=1 * um, n=10)
    morpho.LL = Cylinder(length=5 * um, diameter=2 * um, n=5)
    morpho.right = Cylinder(length=3 * um, diameter=1 * um, n=7)

    assert_equal(morpho.LL.indices[:], [11, 12, 13, 14, 15])
    assert_equal(morpho.L.indices[3 * um : 5 * um], [4, 5])
    assert_equal(
        morpho.L.indices[3 * um : 5 * um], morpho.L[3 * um : 5 * um].indices[:]
    )
    assert_equal(morpho.L.indices[: 5 * um], [1, 2, 3, 4, 5])
    assert_equal(morpho.L.indices[3 * um :], [4, 5, 6, 7, 8, 9, 10])
    assert_equal(morpho.L.indices[3.5 * um], 4)
    assert_equal(morpho.L.indices[3 * um], 4)
    assert_equal(morpho.L.indices[3.9 * um], 4)
    assert_equal(morpho.L.indices[3], 4)
    assert_equal(morpho.L.indices[-1], 10)
    assert_equal(morpho.L.indices[3:5], [4, 5])
    assert_equal(morpho.L.indices[3:], [4, 5, 6, 7, 8, 9, 10])
    assert_equal(morpho.L.indices[:5], [1, 2, 3, 4, 5])


@pytest.mark.codegen_independent
def test_subgroup_attributes():
    morpho = Soma(diameter=30 * um)
    morpho.L = Cylinder(length=10 * um, diameter=1 * um, n=10)
    morpho.LL = Cylinder(x=[0, 5] * um, diameter=2 * um, n=5)
    morpho.right = Cylinder(length=3 * um, diameter=1 * um, n=7)

    # # Getting a single compartment by index
    assert_allclose(morpho.L[2].area, morpho.L.area[2])
    assert_allclose(morpho.L[2].volume, morpho.L.volume[2])
    assert_allclose(morpho.L[2].length, morpho.L.length[2])
    assert_allclose(morpho.L[2].r_length_1, morpho.L.r_length_1[2])
    assert_allclose(morpho.L[2].r_length_2, morpho.L.r_length_2[2])
    assert_allclose(morpho.L[2].distance, morpho.L.distance[2])
    assert_allclose(morpho.L[2].diameter, morpho.L.diameter[2])
    assert morpho.L[2].x is None
    assert morpho.L[2].y is None
    assert morpho.L[2].z is None
    assert morpho.L[2].start_x is None
    assert morpho.L[2].start_y is None
    assert morpho.L[2].start_z is None
    assert morpho.L[2].end_x is None
    assert morpho.L[2].end_y is None
    assert morpho.L[2].end_z is None

    # # Getting a single compartment by position
    assert_allclose(morpho.LL[1.5 * um].area, morpho.LL.area[1])
    assert_allclose(morpho.LL[1.5 * um].volume, morpho.LL.volume[1])
    assert_allclose(morpho.LL[1.5 * um].length, morpho.LL.length[1])
    assert_allclose(morpho.LL[1.5 * um].r_length_1, morpho.LL.r_length_1[1])
    assert_allclose(morpho.LL[1.5 * um].r_length_2, morpho.LL.r_length_2[1])
    assert_allclose(morpho.LL[1.5 * um].distance, morpho.LL.distance[1])
    assert_allclose(morpho.LL[1.5 * um].diameter, morpho.LL.diameter[1])
    assert_allclose(morpho.LL[1.5 * um].x, morpho.LL.x[1])
    assert_allclose(morpho.LL[1.5 * um].y, morpho.LL.y[1])
    assert_allclose(morpho.LL[1.5 * um].z, morpho.LL.z[1])
    assert_allclose(morpho.LL[1.5 * um].start_x, morpho.LL.start_x[1])
    assert_allclose(morpho.LL[1.5 * um].start_y, morpho.LL.start_y[1])
    assert_allclose(morpho.LL[1.5 * um].start_z, morpho.LL.start_z[1])
    assert_allclose(morpho.LL[1.5 * um].end_x, morpho.LL.end_x[1])
    assert_allclose(morpho.LL[1.5 * um].end_y, morpho.LL.end_y[1])
    assert_allclose(morpho.LL[1.5 * um].end_z, morpho.LL.end_z[1])

    # Getting several compartments by indices
    assert_allclose(morpho.right[3:6].area, morpho.right.area[3:6])
    assert_allclose(morpho.right[3:6].volume, morpho.right.volume[3:6])
    assert_allclose(morpho.right[3:6].length, morpho.right.length[3:6])
    assert_allclose(morpho.right[3:6].r_length_1, morpho.right.r_length_1[3:6])
    assert_allclose(morpho.right[3:6].r_length_2, morpho.right.r_length_2[3:6])
    assert_allclose(morpho.right[3:6].distance, morpho.right.distance[3:6])
    assert_allclose(morpho.right[3:6].diameter, morpho.right.diameter[3:6])
    assert morpho.right[3:6].x is None
    assert morpho.right[3:6].y is None
    assert morpho.right[3:6].z is None
    assert morpho.right[3:6].start_x is None
    assert morpho.right[3:6].start_y is None
    assert morpho.right[3:6].start_z is None
    assert morpho.right[3:6].end_x is None
    assert morpho.right[3:6].end_y is None
    assert morpho.right[3:6].end_z is None

    # Getting several compartments by position
    assert_allclose(morpho.L[3 * um : 5 * um].distance, [3.5, 4.5] * um)
    assert_allclose(morpho.L[3.5 * um : 4.5 * um].distance, [3.5, 4.5] * um)


@pytest.mark.codegen_independent
def test_subgroup_incorrect():
    # Incorrect indexing
    morpho = Soma(diameter=30 * um)
    morpho.L = Cylinder(length=10 * um, diameter=1 * um, n=10)
    morpho.LL = Cylinder(length=5 * um, diameter=2 * um, n=5)
    morpho.right = Cylinder(length=3 * um, diameter=1 * um, n=7)

    # Non-existing branch
    with pytest.raises(AttributeError):
        morpho.axon

    # Incorrect indexing
    #  wrong units or mixing units
    with pytest.raises(TypeError):
        morpho.L[3 * second : 5 * second]
    with pytest.raises(TypeError):
        morpho.L[3.4:5.3]
    with pytest.raises(TypeError):
        morpho.L[3 : 5 * um]
    with pytest.raises(TypeError):
        morpho.L[3 * um : 5]
    #   providing a step
    with pytest.raises(TypeError):
        morpho.L[3 * um : 5 * um : 2 * um]
    with pytest.raises(TypeError):
        morpho.L[3:5:2]
    #   incorrect type
    with pytest.raises(TypeError):
        morpho.L[object()]
    # out of range
    with pytest.raises(IndexError):
        morpho.L[-10 * um]
    with pytest.raises(IndexError):
        morpho.L[15 * um]
    with pytest.raises(IndexError):
        morpho.L[10]


@pytest.mark.codegen_independent
def test_topology():
    soma = Soma(diameter=30 * um)
    soma.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    soma.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    soma.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    soma.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    str_topology = str(soma.topology())
    lines = [l for l in str_topology.split("\n") if len(l.strip())]
    assert len(lines) == 5  # one line for each section
    for line, name in zip(lines, ["root", ".L", ".R", ".R.left", "R.right"]):
        assert name in line


@pytest.mark.codegen_independent
def test_copy_section_soma():
    soma = Soma(diameter=30 * um)
    soma_copy = soma.copy_section()
    assert soma_copy.diameter[0] == 30 * um
    assert soma_copy.x is None
    assert soma_copy.y is None
    assert soma_copy.z is None
    assert soma_copy.type == "soma"

    soma = Soma(diameter=30 * um, x=5 * um, z=-10 * um)
    soma_copy = soma.copy_section()
    assert soma_copy.diameter[0] == 30 * um
    assert_allclose(soma_copy.x[0], 5 * um)
    assert_allclose(soma_copy.y[0], 0 * um)
    assert_allclose(soma_copy.z[0], -10 * um)
    assert soma_copy.type == "soma"


@pytest.mark.codegen_independent
def test_copy_section_section():
    # No coordinates
    sec = Section(
        diameter=[10, 5, 4, 3, 2, 1] * um, n=5, length=np.ones(5) * 10 * um, type="dend"
    )
    sec_copy = sec.copy_section()
    assert_allclose(sec_copy.start_diameter, sec.start_diameter)
    assert_allclose(sec_copy.end_diameter, sec.end_diameter)
    assert_allclose(sec_copy.length, sec.length)
    assert sec_copy.n == sec.n
    assert sec_copy.x is None
    assert sec_copy.y is None
    assert sec_copy.z is None
    assert sec_copy.type == "dend"

    # With coordinates
    sec = Section(
        diameter=[10, 5, 4, 3, 2, 1] * um,
        n=5,
        x=[0, 1, 2, 3, 4, 5] * um,
        y=[0, -1, -2, -3, -4, -5] * um,
    )
    sec_copy = sec.copy_section()
    assert_allclose(sec_copy.start_diameter, sec.start_diameter)
    assert_allclose(sec_copy.end_diameter, sec.end_diameter)
    assert_allclose(sec_copy.length, sec.length)
    assert sec_copy.n == sec.n
    assert_allclose(sec_copy.x, sec.x)
    assert_allclose(sec_copy.y, sec.y)
    assert_allclose(sec_copy.z, sec.z)

    assert sec_copy.type is None


@pytest.mark.codegen_independent
def test_copy_section_cylinder():
    # no coordinates
    sec = Section(
        diameter=[10, 5, 4, 3, 2, 1] * um, n=5, length=np.ones(5) * 20 * um, type="dend"
    )
    sec_copy = sec.copy_section()
    assert_allclose(sec_copy.end_diameter, sec.end_diameter)
    assert_allclose(sec_copy.length, sec.length)
    assert sec_copy.n == sec.n
    assert sec_copy.x is None
    assert sec_copy.y is None
    assert sec_copy.z is None
    assert sec_copy.type == "dend"

    # with coordinates
    sec = Section(
        diameter=[10, 5, 4, 3, 2, 1] * um,
        n=5,
        x=[0, 1, 2, 3, 4, 5] * um,
        y=[0, -1, -2, -3, -4, -5] * um,
    )
    sec_copy = sec.copy_section()
    assert_allclose(sec_copy.end_diameter, sec.end_diameter)
    assert_allclose(sec_copy.length, sec.length)
    assert sec_copy.n == sec.n
    assert_allclose(sec_copy.x, sec.x)
    assert_allclose(sec_copy.y, sec.y)
    assert_allclose(sec_copy.z, sec.z)

    assert sec_copy.type is None


def _check_length_coord_consistency(morph_with_coords):
    if not isinstance(morph_with_coords, Soma):
        vectors = np.diff(morph_with_coords.coordinates, axis=0)
        calculated_length = np.sqrt(np.sum(vectors**2, axis=1))
        assert_allclose(calculated_length, morph_with_coords.length)
    for child in morph_with_coords.children:
        _check_length_coord_consistency(child)


@pytest.mark.codegen_independent
def test_generate_coordinates_deterministic():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates()
    assert morph_with_coords.total_compartments == morph.total_compartments
    assert morph_with_coords.total_sections == morph.total_sections

    for new, old in [
        (morph_with_coords, morph),
        (morph_with_coords.L, morph.L),
        (morph_with_coords.R, morph.R),
        (morph_with_coords.R.left, morph.R.left),
        (morph_with_coords.R.right, morph.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)
        # The morphology should be in the x/y plane
        assert_equal(new.z, 0 * um)

    _check_length_coord_consistency(morph_with_coords)


@pytest.mark.codegen_independent
def test_generate_coordinates_random_sections():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates(section_randomness=25)
    assert morph_with_coords.total_compartments == morph.total_compartments
    assert morph_with_coords.total_sections == morph.total_sections

    for new, old in [
        (morph_with_coords, morph),
        (morph_with_coords.L, morph.L),
        (morph_with_coords.R, morph.R),
        (morph_with_coords.R.left, morph.R.left),
        (morph_with_coords.R.right, morph.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)

    _check_length_coord_consistency(morph_with_coords)


@pytest.mark.codegen_independent
def test_generate_coordinates_random_compartments():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates(compartment_randomness=15)
    assert morph_with_coords.total_compartments == morph.total_compartments
    assert morph_with_coords.total_sections == morph.total_sections

    for new, old in [
        (morph_with_coords, morph),
        (morph_with_coords.L, morph.L),
        (morph_with_coords.R, morph.R),
        (morph_with_coords.R.left, morph.R.left),
        (morph_with_coords.R.right, morph.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)

    _check_length_coord_consistency(morph_with_coords)


@pytest.mark.codegen_independent
def test_generate_coordinates_random_all():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates(
        section_randomness=25, compartment_randomness=15
    )
    assert morph_with_coords.total_compartments == morph.total_compartments
    assert morph_with_coords.total_sections == morph.total_sections

    for new, old in [
        (morph_with_coords, morph),
        (morph_with_coords.L, morph.L),
        (morph_with_coords.R, morph.R),
        (morph_with_coords.R.left, morph.R.left),
        (morph_with_coords.R.right, morph.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)

    _check_length_coord_consistency(morph_with_coords)


@pytest.mark.codegen_independent
def test_generate_coordinates_no_overwrite():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates(compartment_randomness=15)
    # This should not change anything because the morphology already has coordinates!
    morph_with_coords2 = morph_with_coords.generate_coordinates(
        section_randomness=25, compartment_randomness=15
    )

    for new, old in [
        (morph_with_coords2, morph_with_coords),
        (morph_with_coords2.L, morph_with_coords.L),
        (morph_with_coords2.R, morph_with_coords.R),
        (morph_with_coords2.R.left, morph_with_coords.R.left),
        (morph_with_coords2.R.right, morph_with_coords.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)
        assert_allclose(new.x, old.x)
        assert_allclose(new.y, old.y)
        assert_allclose(new.z, old.z)


@pytest.mark.codegen_independent
def test_generate_coordinates_overwrite():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    morph_with_coords = morph.generate_coordinates(compartment_randomness=15)
    # This should change things since we explicitly ask for it
    morph_with_coords2 = morph_with_coords.generate_coordinates(
        section_randomness=25, compartment_randomness=15, overwrite_existing=True
    )

    for new, old in [  # ignore the root compartment
        (morph_with_coords2.L, morph_with_coords.L),
        (morph_with_coords2.R, morph_with_coords.R),
        (morph_with_coords2.R.left, morph_with_coords.R.left),
        (morph_with_coords2.R.right, morph_with_coords.R.right),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)
        assert all(np.abs(new.x - old.x) > 0)
        assert all(np.abs(new.y - old.y) > 0)
        assert all(np.abs(new.z - old.z) > 0)

    _check_length_coord_consistency(morph_with_coords2)


@pytest.mark.codegen_independent
def test_generate_coordinates_mixed_overwrite():
    morph = Soma(diameter=30 * um)
    morph.L = Section(
        n=5, diameter=[10, 8, 6, 4, 2, 0] * um, length=np.ones(5) * 20 * um
    )  # tapering truncated cones
    morph.R = Cylinder(n=10, diameter=5 * um, length=50 * um)
    morph_with_coords = morph.generate_coordinates(
        section_randomness=25, compartment_randomness=15
    )
    # The following just returns a copy, as all coordinates are already
    # specified
    morph_copy = morph_with_coords.generate_coordinates()

    # Add new sections that do not yet have coordinates
    morph_with_coords.R.left = Cylinder(n=10, diameter=2.5 * um, length=50 * um)
    morph_with_coords.R.right = Section(
        n=5, diameter=[5, 4, 3, 2, 1, 0] * um, length=np.ones(5) * 10 * um
    )

    # This should change things since we explicitly ask for it
    morph_with_coords2 = morph_with_coords.generate_coordinates(
        section_randomness=25, compartment_randomness=15
    )

    for new, old in [
        (morph_with_coords2, morph_with_coords),
        (morph_with_coords2.L, morph_with_coords.L),
        (morph_with_coords2.R, morph_with_coords.R),
    ]:
        assert new.n == old.n
        assert_allclose(new.length, old.length)
        assert_allclose(new.diameter, old.diameter)
        assert_allclose(new.x, old.x)
        assert_allclose(new.y, old.y)
        assert_allclose(new.z, old.z)

    assert morph_with_coords.R.left.x is None
    assert len(morph_with_coords2.R.left.x) == morph_with_coords2.R.left.n

    _check_length_coord_consistency(morph_with_coords2)


@pytest.mark.codegen_independent
def test_str_repr():
    # A very basic test, make sure that the str/repr functions return
    # something and do not raise an error
    for morph in [
        Soma(diameter=30 * um),
        Soma(diameter=30 * um, x=5 * um, y=10 * um),
        Cylinder(n=5, diameter=10 * um, length=50 * um),
        Cylinder(n=5, diameter=10 * um, x=[0, 50] * um),
        Section(
            n=5, diameter=[2.5, 5, 10, 5, 10, 5] * um, length=[10, 20, 5, 5, 10] * um
        ),
        Section(
            n=5, diameter=[2.5, 5, 10, 5, 10, 5] * um, x=[0, 10, 30, 35, 40, 50] * um
        ),
    ]:
        assert len(repr(morph)) > 0
        assert len(str(morph)) > 0
    morph = Soma(30 * um)
    assert len(repr(morph.children)) > 0
    assert len(str(morph.children)) > 0
    morph.axon = Cylinder(1 * um, n=10, length=100 * um)
    morph.dend = Cylinder(1 * um, n=10, length=50 * um)
    assert len(repr(morph.children)) > 0
    assert len(str(morph.children)) > 0


if __name__ == "__main__":
    test_attributes_soma()
    test_attributes_soma_coordinates()
    test_attributes_cylinder()
    test_attributes_cylinder_coordinates()
    test_attributes_section()
    test_attributes_section_coordinates_single()
    test_attributes_section_coordinates_all()
    test_tree_cables_schematic()
    test_tree_cables_coordinates()
    test_tree_cables_from_points()
    test_tree_cables_from_swc()
    test_tree_soma_schematic()
    test_tree_soma_coordinates()
    test_tree_soma_from_points()
    test_tree_soma_from_points_3_point_soma()
    test_tree_soma_from_points_3_point_soma_incorrect()
    test_tree_soma_from_swc()
    test_tree_soma_from_swc_3_point_soma()
    test_construction_incorrect_arguments()
    test_from_points_minimal()
    test_from_points_incorrect()
    test_subtree_deletion()
    test_subgroup_indices()
    test_subgroup_attributes()
    test_subgroup_incorrect()
    test_topology()
    test_copy_section_soma()
    test_copy_section_section()
    test_copy_section_cylinder()
    test_generate_coordinates_deterministic()
    test_generate_coordinates_random_sections()
    test_generate_coordinates_random_compartments()
    test_generate_coordinates_random_all()
    test_generate_coordinates_no_overwrite()
    test_generate_coordinates_overwrite()
    test_generate_coordinates_mixed_overwrite()
    test_str_repr()



#!/usr/bin/env python
# coding: utf-8
"""
[Jansen and Rit 1995 model](https://link.springer.com/content/pdf/10.1007/BF00199471.pdf) (Figure 3) in Brian2.

Equations are the system of differential equations number (6) in the original paper. 
The rate parameters $a=100 s^{-1}$ and $b=200 s^{-1}$ were changed to excitatory $\tau_e = 1000ms/a =10ms$  and inhibitory $\tau_i = 1000ms/b =20ms$ time constants as in 
[Thomas Knosche review](https://link.springer.com/referenceworkentry/10.1007%2F978-1-4614-6675-8_65), 
[Touboul et al. 2011](https://direct.mit.edu/neco/article-abstract/23/12/3232/7717/Neural-Mass-Activity-Bifurcations-and-Epilepsy?redirectedFrom=fulltext), or 
[David & Friston 2003](https://www.sciencedirect.com/science/article/pii/S1053811903004579). 
Units were removerd from parameters $e_0$, $v_0$, $r_0$, $A$, $B$, and $p$ to stop Brian's confusion.

Ruben Tikidji-Hamburyan 2021 (rth@r-a-r.org)
""" 



from numpy import *
from numpy import random as rnd
from matplotlib.pyplot import *
from brian2 import *

defaultclock.dt = .1*ms   #default time step

te,ti    = 10.*ms, 20.*ms #taus for excitatory and inhibitory populations
e0       = 5.             #max firing rate
v0       = 6.             #(max FR)/2 input
r0       = 0.56           #gain rate
A,B,C    = 3.25, 22., 135 #standard parameters as in the set (7) of the original paper
P,deltaP = 120, 320.-120  #random input uniformly distributed between 120 and
                          #320 pulses per second

# Random noise
nstim = TimedArray(rnd.rand(70000),2*ms)

# Equations as in the system (6) of the original paper
equs = """
dy0/dt = y3                             /second : 1
dy3/dt = (A        * Sp -2*y3 -y0/te*second)/te : 1
dy1/dt = y4                             /second : 1
dy4/dt = (A*(p+ C2 * Se)-2*y4 -y1/te*second)/te : 1
dy2/dt = y5                             /second : 1
dy5/dt = (B   * C4 * Si -2*y5 -y2/ti*second)/ti : 1
p  = P0+nstim(t)*dlP                   : 1
Sp = e0/(1+exp(r0*(v0 - (y1-y2)    ))) : 1
Se = e0/(1+exp(r0*(v0 -  C1*y0     ))) : 1
Si = e0/(1+exp(r0*(v0 -  C3*y0     ))) : 1
C1            : 1
C2 = 0.8 *C1  : 1
C3 = 0.25*C1  : 1
C4 = 0.25*C1  : 1
P0            : 1
dlP           : 1
"""

n = NeuronGroup(6,equs,method='euler') #creates 6 JR models for different connectivity parameters

#set parameters as for different traces on figure 3 of the original paper
n.C1[0]  = 68
n.C1[1]  = 128
n.C1[2]  = C
n.C1[3]  = 270
n.C1[4]  = 675
n.C1[5]  = 1350
#set stimulus offset and noise magnitude
n.P0     = P
n.dlP    = deltaP

#just record everything
sm = StateMonitor(n,['y4','y1','y3','y0','y5','y2'],record=True)

#Runs for 5 second
run(5*second,report='text')


#This code goes over all models with different parameters and plot activity of each population.

figure(1,figsize=(22,16))
idx1 = where(sm.t/second>2.)[0]

o = 0
for p in [0,1,2,3,4,5]:
    if o == 0: ax = subplot(6,3,1)
    else     :subplot(6,3,1+o,sharex=ax)
    if o == 0: title("E")
    plot(sm.t[idx1]/second, sm[p].y1[idx1],'g-')
    ylabel(f"C={n[p].C1[0]}")
    if o == 15: xlabel("Time (seconds)")
    subplot(6,3,2+o,sharex=ax)
    if o == 0: title("P")
    plot(sm.t[idx1]/second, sm[p].y0[idx1],'b-')
    if o == 15: xlabel("Time (seconds)")
    subplot(6,3,3+o,sharex=ax)
    if o == 0: title("I")
    plot(sm.t[idx1]/second, sm[p].y2[idx1],'r-')
    if o == 15: xlabel("Time (seconds)")
    o += 3

show()




# encoding: utf8
"""
Morris-Lecar model

Reproduces Fig. 9 of:

Catherine Morris and Harold Lecar.
“Voltage Oscillations in the Barnacle Giant Muscle Fiber.”
Biophysical Journal 35, no. 1 (1981): 193–213.
"""
from brian2 import *
set_device('cpp_standalone')
defaultclock.dt = 0.01*ms

g_L = 2*mS
g_Ca = 4*mS
g_K = 8*mS
V_L = -50*mV
V_Ca = 100*mV
V_K = -70*mV
lambda_n__max = 1.0/(15*ms)
V_1 = 10*mV
V_2 = 15*mV  # Note that Figure caption says -15 which seems to be a typo
V_3 = -1*mV
V_4 = 14.5*mV
C = 20*uF


# V,N-reduced system (Eq. 9 in article), note that the variables M and N (and lambda_N, etc.)
# have been renamed to m and n to better match the Hodgkin-Huxley convention, and because N has
# a reserved meaning in Brian (number of neurons)
eqs = '''
dV/dt = (-g_L*(V - V_L) - g_Ca*m_inf*(V - V_Ca) - g_K*n*(V - V_K) + I)/C : volt
dn/dt = lambda_n*(n_inf - n) : 1
m_inf = 0.5*(1 + tanh((V - V_1)/V_2)) : 1
n_inf = 0.5*(1 + tanh((V - V_3)/V_4)) : 1
lambda_n = lambda_n__max*cosh((V - V_3)/(2*V_4)) : Hz
I : amp
'''

neuron = NeuronGroup(17, eqs, method='exponential_euler')
neuron.I = (np.arange(17)*25+100)*uA
neuron.V = V_L
neuron.n = 'n_inf'
mon = StateMonitor(neuron, ['V', 'n'], record=True)

run_time = 220*ms
run(run_time)
fig, (ax1, ax2) = plt.subplots(1, 2, gridspec_kw={'right': 0.95, 'bottom': 0.15},
                               figsize=(6.4, 3.2))
fig.subplots_adjust(wspace=0.4)
for line_no, idx in enumerate([0, 4, 12, 15]):
    color = 'C%d' % line_no
    ax1.plot(mon.t/ms, mon.V[idx]/mV, color=color)
    ax1.text(225, mon.V[idx][-1]/mV, '%.0f' % (neuron.I[idx]/uA), color=color)
ax1.set(xlim=(0, 220), ylim=(-50, 50), xlabel='time (ms)')
ax1.set_ylabel('V (mV)', rotation=0)
ax1.spines['right'].set_visible(False)
ax1.spines['top'].set_visible(False)

# dV/dt nullclines
V = linspace(-50, 50, 100)*mV
for line_no, (idx, color) in enumerate([(0, 'C0'), (4, 'C1'), (8, 'C4'), (12, 'C2'), (16, 'C5')]):
    n_null = (g_L*(V - V_L) + g_Ca*0.5*(1 + tanh((V - V_1)/V_2))*(V - V_Ca) - neuron.I[idx])/(-g_K*(V - V_K))
    ax2.plot(V/mV, n_null, color=color)
    ax2.text(V[20+5*line_no]/mV, n_null[20+5*line_no]+0.01, '%.0f' % (neuron.I[idx]/uA), color=color)
# dn/dt nullcline
n_null = 0.5*(1 + tanh((V - V_3)/V_4))
ax2.plot(V/mV, n_null, color='k')
ax2.set(xlim=(-50, 50), ylim=(0, 1), xlabel='V (mV)')
ax2.set_ylabel('n', rotation=0)
ax2.spines['right'].set_visible(False)
ax2.spines['top'].set_visible(False)
plt.show()



"""
A simple example showing how string expressions can be used to implement
spatial (deterministic or stochastic) connection patterns.
"""
from brian2 import *

rows, cols = 20, 20
G = NeuronGroup(rows * cols, '''x : meter
                                y : meter''')
# initialize the grid positions
grid_dist = 25*umeter
G.x = '(i // rows) * grid_dist - rows/2.0 * grid_dist'
G.y = '(i % rows) * grid_dist - cols/2.0 * grid_dist'

# Deterministic connections
distance = 120*umeter
S_deterministic = Synapses(G, G)
S_deterministic.connect('sqrt((x_pre - x_post)**2 + (y_pre - y_post)**2) < distance')

# Random connections (no self-connections)
S_stochastic = Synapses(G, G)
S_stochastic.connect('i != j',
                     p='1.5 * exp(-((x_pre-x_post)**2 + (y_pre-y_post)**2)/(2*(60*umeter)**2))')

figure(figsize=(12, 6))

# Show the connections for some neurons in different colors
for color in ['g', 'b', 'm']:
    subplot(1, 2, 1)
    neuron_idx = np.random.randint(0, rows*cols)
    plot(G.x[neuron_idx] / umeter, G.y[neuron_idx] / umeter, 'o', mec=color,
         mfc='none')
    plot(G.x[S_deterministic.j[neuron_idx, :]] / umeter,
         G.y[S_deterministic.j[neuron_idx, :]] / umeter, color + '.')
    subplot(1, 2, 2)
    plot(G.x[neuron_idx] / umeter, G.y[neuron_idx] / umeter, 'o', mec=color,
         mfc='none')
    plot(G.x[S_stochastic.j[neuron_idx, :]] / umeter,
         G.y[S_stochastic.j[neuron_idx, :]] / umeter, color + '.')

for idx, t in enumerate(['determininstic connections',
                         'random connections']):
    subplot(1, 2, idx + 1)
    xlim((-rows/2.0 * grid_dist) / umeter, (rows/2.0 * grid_dist) / umeter)
    ylim((-cols/2.0 * grid_dist) / umeter, (cols/2.0 * grid_dist) / umeter)
    title(t)
    xlabel('x')
    ylabel('y', rotation='horizontal')
    axis('equal')

tight_layout()
show()



import numpy as np
import pytest
from numpy.testing import assert_equal

from brian2.memory.dynamicarray import DynamicArray, DynamicArray1D


@pytest.mark.codegen_independent
def test_dynamic_array_1d_access():
    da = DynamicArray1D(10)
    da[:] = np.arange(10)
    assert da[7] == 7
    assert len(da) == 10
    assert da.shape == (10,)
    assert len(str(da))
    assert len(repr(da))
    da[:] += 1
    da.data[:] += 1
    assert all(da[:] == (np.arange(10) + 2))


@pytest.mark.codegen_independent
def test_dynamic_array_1d_resize_up_down():
    for numpy_resize in [True, False]:
        da = DynamicArray1D(10, use_numpy_resize=numpy_resize, refcheck=False)
        da[:] = np.arange(10)
        da.resize(15)
        assert len(da) == 15
        assert da.shape == (15,)
        assert all(da[10:] == 0)
        assert all(da[:10] == np.arange(10))
        da.resize(5)
        assert len(da) == 5
        assert da.shape == (5,)
        assert all(da[:] == np.arange(5))


@pytest.mark.codegen_independent
def test_dynamic_array_1d_resize_down_up():
    for numpy_resize in [True, False]:
        da = DynamicArray1D(10, use_numpy_resize=numpy_resize)
        da[:] = np.arange(10)
        da.resize(5)
        assert len(da) == 5
        assert da.shape == (5,)
        assert all(da[:5] == np.arange(5))
        da.resize(10)
        assert len(da) == 10
        assert da.shape == (10,)
        assert all(da[:5] == np.arange(5))
        assert all(da[5:] == 0)


@pytest.mark.codegen_independent
def test_dynamic_array_1d_shrink():
    for numpy_resize in [True, False]:
        da = DynamicArray1D(10, use_numpy_resize=numpy_resize, refcheck=False)
        da[:] = np.arange(10)
        da.shrink(5)
        assert len(da) == 5
        assert all(da[:] == np.arange(5))
        # After using shrink, the underlying array should have changed
        assert len(da._data) == 5


@pytest.mark.codegen_independent
def test_dynamic_array_2d_access():
    da = DynamicArray1D((10, 20))
    da[:, :] = np.arange(200).reshape((10, 20))
    assert da[5, 10] == 5 * 20 + 10
    assert da.shape == (10, 20)
    assert len(str(da))
    assert len(repr(da))
    da[:] += 1
    da.data[:] += 1
    assert_equal(da[:, :], np.arange(200).reshape((10, 20)) + 2)


@pytest.mark.codegen_independent
def test_dynamic_array_2d_resize_up_down():
    for numpy_resize in [True, False]:
        da = DynamicArray((10, 20), use_numpy_resize=numpy_resize, refcheck=False)
        da[:, :] = np.arange(200).reshape((10, 20))
        da.resize((15, 20))
        assert da.shape == (15, 20)
        assert_equal(da[10:, :], np.zeros((5, 20)))
        assert_equal(da[:10, :], np.arange(200).reshape((10, 20)))
        da.resize((15, 25))
        assert da.shape == (15, 25)
        assert_equal(da[:10, 20:], np.zeros((10, 5)))
        assert_equal(da[:10, :20], np.arange(200).reshape((10, 20)))

        da.resize((10, 20))
        assert da.shape == (10, 20)
        assert_equal(da[:, :], np.arange(200).reshape((10, 20)))


@pytest.mark.codegen_independent
def test_dynamic_array_2d_resize_down_up():
    for numpy_resize in [True, False]:
        da = DynamicArray((10, 20), use_numpy_resize=numpy_resize, refcheck=False)
        da[:, :] = np.arange(200).reshape((10, 20))
        da.resize((5, 20))
        assert da.shape == (5, 20)
        assert_equal(da, np.arange(100).reshape((5, 20)))
        da.resize((5, 15))
        assert da.shape == (5, 15)
        for row_idx, row in enumerate(da):
            assert_equal(row, 20 * row_idx + np.arange(15))

        da.resize((10, 20))
        assert da.shape == (10, 20)
        for row_idx, row in enumerate(da[:5, :15]):
            assert_equal(row, 20 * row_idx + np.arange(15))
        assert_equal(da[5:, 15:], 0)


@pytest.mark.codegen_independent
def test_dynamic_array_2d_shrink():
    for numpy_resize in [True, False]:
        da = DynamicArray((10, 20), use_numpy_resize=numpy_resize, refcheck=False)
        da[:, :] = np.arange(200).reshape((10, 20))
        da.shrink((5, 15))
        assert da.shape == (5, 15)
        # After using shrink, the underlying array should have changed
        assert da._data.shape == (5, 15)
        assert_equal(
            da[:, :], np.arange(15).reshape((1, 15)) + 20 * np.arange(5).reshape((5, 1))
        )


if __name__ == "__main__":
    test_dynamic_array_1d_access()
    test_dynamic_array_1d_resize_up_down()
    test_dynamic_array_1d_resize_down_up()
    test_dynamic_array_1d_shrink()
    test_dynamic_array_2d_access()
    test_dynamic_array_2d_resize_up_down()
    test_dynamic_array_2d_resize_down_up()
    test_dynamic_array_2d_shrink()



"""
A model with adaptive threshold (increases with each spike)
"""
from brian2 import *

eqs = '''
dv/dt = -v/(10*ms) : volt
dvt/dt = (10*mV-vt)/(15*ms) : volt
'''

reset = '''
v = 0*mV
vt += 3*mV
'''

IF = NeuronGroup(1, model=eqs, reset=reset, threshold='v>vt',
                 method='exact')
IF.vt = 10*mV
PG = PoissonGroup(1, 500 * Hz)

C = Synapses(PG, IF, on_pre='v += 3*mV')
C.connect()

Mv = StateMonitor(IF, 'v', record=True)
Mvt = StateMonitor(IF, 'vt', record=True)
# Record the value of v when the threshold is crossed
M_crossings = SpikeMonitor(IF, variables='v')
run(2*second, report='text')

subplot(1, 2, 1)
plot(Mv.t / ms, Mv[0].v / mV)
plot(Mvt.t / ms, Mvt[0].vt / mV)
ylabel('v (mV)')
xlabel('t (ms)')
# zoom in on the first 100ms
xlim(0, 100)
subplot(1, 2, 2)
hist(M_crossings.v / mV, bins=np.arange(10, 20, 0.5))
xlabel('v at threshold crossing (mV)')
show()



import brian2
import sphinx
import doc2dash
import sys
import os

# - Build documentation with Sphinx
os.chdir('../../../docs_sphinx')
sphinx.main(['sphinx-build', '-b', 'html', '-d', '_build/doctrees', '.', '_build/html'])

# - Run doc2dash on the built documentation
os.system('doc2dash _build/html/ -n Brian2 -I index.html -d ..')




# coding=utf-8
"""
Modeling neuron-glia interactions with the Brian 2 simulator
Marcel Stimberg, Dan F. M. Goodman, Romain Brette, Maurizio De Pittà
bioRxiv 198366; doi: https://doi.org/10.1101/198366

Figure 5: Astrocytes connected in a network.

Intercellular calcium wave propagation in a ring of 50 astrocytes connected by
bidirectional gap junctions (see Goldberg et al., 2010)
"""
from brian2 import *

import plot_utils as pu

set_device('cpp_standalone', directory=None)  # Use fast "C++ standalone mode"

################################################################################
# Model parameters
################################################################################
### General parameters
duration = 4000*second       # Total simulation time
sim_dt = 50*ms               # Integrator/sampling step

### Astrocyte parameters
# ---  Calcium fluxes
O_P = 0.9*umolar/second      # Maximal Ca^2+ uptake rate by SERCAs
K_P = 0.05 * umolar          # Ca2+ affinity of SERCAs
C_T = 2*umolar               # Total cell free Ca^2+ content
rho_A = 0.18                 # ER-to-cytoplasm volume ratio
Omega_C = 6/second           # Maximal rate of Ca^2+ release by IP_3Rs
Omega_L = 0.1/second         # Maximal rate of Ca^2+ leak from the ER
# --- IP_3R kinectics
d_1 = 0.13*umolar            # IP_3 binding affinity
d_2 = 1.05*umolar            # Ca^2+ inactivation dissociation constant
O_2 = 0.2/umolar/second      # IP_3R binding rate for Ca^2+ inhibition
d_3 = 0.9434*umolar          # IP_3 dissociation constant
d_5 = 0.08*umolar            # Ca^2+ activation dissociation constant
# --- IP_3 production
O_delta = 0.6*umolar/second  # Maximal rate of IP_3 production by PLCdelta
kappa_delta = 1.5* umolar    # Inhibition constant of PLC_delta by IP_3
K_delta = 0.1*umolar         # Ca^2+ affinity of PLCdelta
# --- IP_3 degradation
Omega_5P = 0.05/second       # Maximal rate of IP_3 degradation by IP-5P
K_D = 0.7*umolar             # Ca^2+ affinity of IP3-3K
K_3K = 1.0*umolar            # IP_3 affinity of IP_3-3K
O_3K = 4.5*umolar/second     # Maximal rate of IP_3 degradation by IP_3-3K
# --- IP_3 diffusion
F_ex = 0.09*umolar/second    # Maximal exogenous IP3 flow
F = 0.09*umolar/second       # GJC IP_3 permeability
I_Theta = 0.3*umolar         # Threshold gradient for IP_3 diffusion
omega_I = 0.05*umolar        # Scaling factor of diffusion

################################################################################
# Model definition
################################################################################
defaultclock.dt = sim_dt     # Set the integration time

### Astrocytes
astro_eqs = '''
dI/dt = J_delta - J_3K - J_5P + J_ex + J_coupling : mmolar
J_delta = O_delta/(1 + I/kappa_delta) * C**2/(C**2 + K_delta**2) : mmolar/second
J_3K = O_3K * C**4/(C**4 + K_D**4) * I/(I + K_3K)                : mmolar/second
J_5P = Omega_5P*I                                                : mmolar/second
# Exogenous stimulation (rectangular wave with period of 50s and duty factor 0.4)
stimulus = int((t % (50*second))<20*second)                      : 1
delta_I_bias = I - I_bias*stimulus                               : mmolar
J_ex = -F_ex/2*(1 + tanh((abs(delta_I_bias) - I_Theta)/omega_I)) *
                sign(delta_I_bias)                               : mmolar/second
# Diffusion between astrocytes
J_coupling : mmolar/second

# Ca^2+-induced Ca^2+ release:
dC/dt = J_r + J_l - J_p                                   : mmolar
dh/dt = (h_inf - h)/tau_h                                 : 1
J_r = (Omega_C * m_inf**3 * h**3) * (C_T - (1 + rho_A)*C) : mmolar/second
J_l = Omega_L * (C_T - (1 + rho_A)*C)                     : mmolar/second
J_p = O_P * C**2/(C**2 + K_P**2)                          : mmolar/second
m_inf = I/(I + d_1) * C/(C + d_5)                         : 1
h_inf = Q_2/(Q_2 + C)                                     : 1
tau_h = 1/(O_2 * (Q_2 + C))                               : second
Q_2 = d_2 * (I + d_1)/(I + d_3)                           : mmolar

# External IP_3 drive
I_bias : mmolar (constant)
'''

N_astro = 50 # Total number of astrocytes in the network
astrocytes = NeuronGroup(N_astro, astro_eqs, method='rk4')
# Asymmetric stimulation on the 50th cell to get some nice chaotic patterns
astrocytes.I_bias[N_astro//2] = 1.0*umolar
astrocytes.h = 0.9
# Diffusion between astrocytes
astro_to_astro_eqs = '''
delta_I = I_post - I_pre        : mmolar
J_coupling_post = -F/2 * (1 + tanh((abs(delta_I) - I_Theta)/omega_I)) *
                  sign(delta_I) : mmolar/second (summed)
'''
astro_to_astro = Synapses(astrocytes, astrocytes,
                          model=astro_to_astro_eqs)
# Couple neighboring astrocytes: two connections per astrocyte pair, as
# the above formulation will only update the I_coupling term of one of the
# astrocytes
astro_to_astro.connect('j == (i + 1) % N_pre or '
                       'j == (i + N_pre - 1) % N_pre')

################################################################################
# Monitors
################################################################################
astro_mon = StateMonitor(astrocytes, variables=['C'], record=True)

################################################################################
# Simulation run
################################################################################
run(duration, report='text')

################################################################################
# Analysis and plotting
################################################################################
plt.style.use('figures.mplstyle')

fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6.26894, 6.26894 * 0.66),
                       gridspec_kw={'left': 0.1, 'bottom': 0.12})
scaling = 1.2
step = 10
ax.plot(astro_mon.t/second,
        (astro_mon.C[0:N_astro//2-1].T/astro_mon.C.max() +
         np.arange(N_astro//2-1)*scaling), color='black')
ax.plot(astro_mon.t/second, (astro_mon.C[N_astro//2:].T/astro_mon.C.max() +
                             np.arange(N_astro//2, N_astro)*scaling),
        color='black')
ax.plot(astro_mon.t/second, (astro_mon.C[N_astro//2-1].T/astro_mon.C.max() +
                             np.arange(N_astro//2-1, N_astro//2)*scaling),
        color='C0')
ax.set(xlim=(0., duration/second), ylim=(0, (N_astro+1.5)*scaling),
       xticks=np.arange(0., duration/second, 500), xlabel='time (s)',
       yticks=np.arange(0.5*scaling, (N_astro + 1.5)*scaling, step*scaling),
       yticklabels=[str(yt) for yt in np.arange(0, N_astro + 1, step)],
       ylabel='$C/C_{max}$ (cell index)')
pu.adjust_spines(ax, ['left', 'bottom'])

pu.adjust_ylabels([ax], x_offset=-0.08)

plt.show()


"""
Helper module containing functions that operate on numpy arrays.
"""

import numpy as np


def calc_repeats(delay):
    """
    Calculates offsets corresponding to an array, where repeated values are
    subsequently numbered, i.e. if there n identical values, the returned array
    will have values from 0 to n-1 at their positions.
    The code is complex because tricks are needed for vectorisation.

    This function is used in the Python `SpikeQueue` to calculate the offset
    array for the insertion of spikes with their respective delays into the
    queue and in the numpy code for synapse creation to calculate how many
    synapses for each source-target pair exist.

    Examples
    --------
    >>> import numpy as np
    >>> print(calc_repeats(np.array([7, 5, 7, 3, 7, 5])))
    [0 0 1 0 2 1]
    """
    # We use merge sort because it preserves the input order of equal
    # elements in the sorted output
    sort_indices = np.argsort(delay, kind="mergesort")
    xs = delay[sort_indices]
    J = xs[1:] != xs[:-1]
    A = np.hstack((0, np.cumsum(J)))
    B = np.hstack((0, np.cumsum(np.logical_not(J))))
    BJ = np.hstack((0, B[:-1][J]))
    ei = B - BJ[A]
    ofs = np.zeros_like(delay, dtype=np.int32)
    ofs[sort_indices] = np.array(ei, dtype=ofs.dtype)
    return ofs



"""
Base class for generating code in different programming languages, gives the
methods which should be overridden to implement a new language.
"""

from brian2.codegen.permutation_analysis import (
    OrderDependenceError,
    check_for_order_independence,
)
from brian2.codegen.translation import make_statements
from brian2.core.functions import Function
from brian2.core.variables import ArrayVariable
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers

__all__ = ["CodeGenerator"]


logger = get_logger(__name__)


class CodeGenerator:
    """
    Base class for all languages.

    See definition of methods below.

    TODO: more details here
    """

    # Subclasses should override this
    class_name = ""

    def __init__(
        self,
        variables,
        variable_indices,
        owner,
        iterate_all,
        codeobj_class,
        name,
        template_name,
        override_conditional_write=None,
        allows_scalar_write=False,
    ):
        # We have to do the import here to avoid circular import dependencies.
        from brian2.devices.device import get_device

        self.device = get_device()
        self.variables = variables
        self.variable_indices = variable_indices
        self.func_name_replacements = {}
        for varname, var in variables.items():
            if isinstance(var, Function):
                if codeobj_class in var.implementations:
                    impl_name = var.implementations[codeobj_class].name
                    if impl_name is not None:
                        self.func_name_replacements[varname] = impl_name
        self.iterate_all = iterate_all
        self.codeobj_class = codeobj_class
        self.owner = owner
        if override_conditional_write is None:
            self.override_conditional_write = set()
        else:
            self.override_conditional_write = set(override_conditional_write)
        self.allows_scalar_write = allows_scalar_write
        self.name = name
        self.template_name = template_name
        # Gather the names of functions that should get an additional
        # "_vectorisation_idx" argument in the generated code. Take care
        # of storing their translated name (e.g. "_rand" instead of "rand")
        # if necessary
        self.auto_vectorise = {
            self.func_name_replacements.get(name, name)
            for name in self.variables
            if getattr(self.variables[name], "auto_vectorise", False)
        }

    @staticmethod
    def get_array_name(var, access_data=True):
        """
        Get a globally unique name for a `ArrayVariable`.

        Parameters
        ----------
        var : `ArrayVariable`
            The variable for which a name should be found.
        access_data : bool, optional
            For `DynamicArrayVariable` objects, specifying `True` here means the
            name for the underlying data is returned. If specifying `False`,
            the name of object itself is returned (e.g. to allow resizing).
        Returns
        -------
        name : str
            A uniqe name for `var`.
        """
        # We have to do the import here to avoid circular import dependencies.
        from brian2.devices.device import get_device

        device = get_device()
        return device.get_array_name(var, access_data=access_data)

    def translate_expression(self, expr):
        """
        Translate the given expression string into a string in the target
        language, returns a string.
        """
        raise NotImplementedError

    def translate_statement(self, statement):
        """
        Translate a single line `Statement` into the target language, returns
        a string.
        """
        raise NotImplementedError

    def determine_keywords(self):
        """
        A dictionary of values that is made available to the templated. This is
        used for example by the `CPPCodeGenerator` to set up all the supporting
        code
        """
        return {}

    def translate_one_statement_sequence(self, statements, scalar=False):
        raise NotImplementedError

    def translate_statement_sequence(self, scalar_statements, vector_statements):
        """
        Translate a sequence of `Statement` into the target language, taking
        care to declare variables, etc. if necessary.

        Returns a tuple ``(scalar_code, vector_code, kwds)`` where
        ``scalar_code`` is a list of the lines of code executed before the inner
        loop, ``vector_code`` is a list of the lines of code in the inner
        loop, and ``kwds`` is a dictionary of values that is made available to
        the template.
        """
        scalar_code = {}
        vector_code = {}
        for name, block in scalar_statements.items():
            scalar_code[name] = self.translate_one_statement_sequence(
                block, scalar=True
            )
        for name, block in vector_statements.items():
            vector_code[name] = self.translate_one_statement_sequence(
                block, scalar=False
            )

        kwds = self.determine_keywords()

        return scalar_code, vector_code, kwds

    def array_read_write(self, statements):
        """
        Helper function, gives the set of ArrayVariables that are read from and
        written to in the series of statements. Returns the pair read, write
        of sets of variable names.
        """
        variables = self.variables
        variable_indices = self.variable_indices
        read = set()
        write = set()
        for stmt in statements:
            ids = get_identifiers(stmt.expr)
            # if the operation is inplace this counts as a read.
            if stmt.inplace:
                ids.add(stmt.var)
            read = read.union(ids)
            if stmt.scalar or variable_indices[stmt.var] == "0":
                if stmt.op != ":=" and not self.allows_scalar_write:
                    raise SyntaxError(
                        f"Writing to scalar variable {stmt.var} not allowed in this"
                        " context."
                    )
                for name in ids:
                    if (
                        name in variables
                        and isinstance(variables[name], ArrayVariable)
                        and not (
                            variables[name].scalar or variable_indices[name] == "0"
                        )
                    ):
                        raise SyntaxError(
                            "Cannot write to scalar variable "
                            f"'{stmt.var}' with an expression "
                            f"referring to vector variable '{name}'"
                        )
            write.add(stmt.var)
        read = {
            varname
            for varname, var in list(variables.items())
            if isinstance(var, ArrayVariable) and varname in read
        }
        write = {
            varname
            for varname, var in list(variables.items())
            if isinstance(var, ArrayVariable) and varname in write
        }
        # Gather the indices stored as arrays (ignore _idx which is special)
        indices = set()
        indices |= {
            variable_indices[varname]
            for varname in read
            if not variable_indices[varname] in ("_idx", "0")
            and isinstance(variables[variable_indices[varname]], ArrayVariable)
        }
        indices |= {
            variable_indices[varname]
            for varname in write
            if not variable_indices[varname] in ("_idx", "0")
            and isinstance(variables[variable_indices[varname]], ArrayVariable)
        }
        # don't list arrays that are read explicitly and used as indices twice
        read -= indices
        return read, write, indices

    def get_conditional_write_vars(self):
        """
        Helper function, returns a dict of mappings ``(varname, condition_var_name)`` indicating that
        when ``varname`` is written to, it should only be when ``condition_var_name`` is ``True``.
        """
        conditional_write_vars = {}
        for varname, var in list(self.variables.items()):
            if getattr(var, "conditional_write", None) is not None:
                cvar = var.conditional_write
                cname = cvar.name
                if cname not in self.override_conditional_write:
                    conditional_write_vars[varname] = cname
        return conditional_write_vars

    def arrays_helper(self, statements):
        """
        Combines the two helper functions `array_read_write` and `get_conditional_write_vars`, and updates the
        ``read`` set.
        """
        read, write, indices = self.array_read_write(statements)
        conditional_write_vars = self.get_conditional_write_vars()
        read |= {var for var in write if var in conditional_write_vars}
        read |= {
            conditional_write_vars[var]
            for var in write
            if var in conditional_write_vars
        }
        return read, write, indices, conditional_write_vars

    def has_repeated_indices(self, statements):
        """
        Whether any of the statements potentially uses repeated indices (e.g.
        pre- or postsynaptic indices).
        """
        variables = self.variables
        variable_indices = self.variable_indices
        read, write, indices, conditional_write_vars = self.arrays_helper(statements)
        # Check whether we potentially deal with repeated indices (which will
        # be the case most importantly when we write to pre- or post-synaptic
        # variables in synaptic code)
        used_indices = {variable_indices[var] for var in write}
        all_unique = all(
            variables[index].unique
            for index in used_indices
            if index not in ("_idx", "0")
        )
        return not all_unique

    def translate(self, code, dtype):
        """
        Translates an abstract code block into the target language.
        """
        scalar_statements = {}
        vector_statements = {}
        for ac_name, ac_code in code.items():
            statements = make_statements(
                ac_code, self.variables, dtype, optimise=True, blockname=ac_name
            )
            scalar_statements[ac_name], vector_statements[ac_name] = statements
        for vs in vector_statements.values():
            # Check that the statements are meaningful independent on the order of
            # execution (e.g. for synapses)
            try:
                if self.has_repeated_indices(
                    vs
                ):  # only do order dependence if there are repeated indices
                    check_for_order_independence(
                        vs, self.variables, self.variable_indices
                    )
            except OrderDependenceError:
                # If the abstract code is only one line, display it in full
                if len(vs) <= 1:
                    error_msg = f"Abstract code: '{vs[0]}'\n"
                else:
                    error_msg = (
                        f"{len(vs)} lines of abstract code, first line is: '{vs[0]}'\n"
                    )
                logger.warn(
                    "Came across an abstract code block that may not be "
                    "well-defined: the outcome may depend on the "
                    "order of execution. You can ignore this warning if "
                    "you are sure that the order of operations does not "
                    "matter. " + error_msg
                )

        translated = self.translate_statement_sequence(
            scalar_statements, vector_statements
        )

        return translated



"""
Coupled oscillators, following the Kuramoto model. The current state of an oscillator
is given by its phase :math:`\Theta`, which follows

.. math::

  \frac{d\Theta_i}{dt} = \omega_i + \frac{K}{N}\sum_j sin(\Theta_j - \Theta_i)

where :math:`\omega_i` is the intrinsic frequency of each oscillator, :math:`K` is
the coupling strength, and the sum is over all oscillators (all-to-all coupling).

The plots show a dot on the unit circle denoting the phase of each neuron (with the
color representing the initial phase at the start of the simulation). The black dot and
line show the average phase (dot) and the phase coherence (length of the line). The
simulations are run four times with different coupling strengths :math:`K`, each
simulation starting from the same initial phase distribution.

https://en.wikipedia.org/wiki/Kuramoto_model
"""
import matplotlib.animation as animation

from brian2 import *
from brian2.core.functions import timestep

### global parameters
N = 100
defaultclock.dt = 1*ms


### simulation code
def run_sim(K, random_seed=214040893):
    seed(random_seed)

    eqs = '''
    dTheta/dt = omega + K/N*coupling : radian
    omega : radian/second (constant) # intrinsic frequency
    coupling : 1
    '''

    oscillators = NeuronGroup(N, eqs, method='euler')
    oscillators.Theta = 'rand()*2*pi'  # random initial phase
    oscillators.omega = 'clip(0.5 + randn()*0.5, 0, inf)*radian/second'  # 𝒩(0.5, 0.5)

    connections = Synapses(oscillators, oscillators,
                        'coupling_post = sin(Theta_pre - Theta_post) : 1 (summed)')
    connections.connect()  # all-to-all

    mon = StateMonitor(oscillators, 'Theta', record=True)
    run(10*second)
    return mon.Theta[:]

### Create animated plots
frame_delay = 40*ms

# Helper functions
def to_x_y(phases):
    return np.cos(phases), np.sin(phases)

def calc_coherence_and_phase(x, y):
    phi = np.arctan2(np.sum(y), np.sum(x))
    r = np.sqrt(np.sum(x)**2 + np.sum(y)**2)/N
    return r, phi

# Plot an animation with the phase of each oscillator and the average phase
def do_animation(fig, axes, K_values, theta_values):
    '''
    Makes animated subplots in the given ``axes``, where each ``theta_values`` entry
    is the full recording of ``Theta`` from the monitor.
    '''
    artists = []
    for ax, K, Theta in zip(axes, K_values, theta_values):
        x, y = to_x_y(Theta.T[0])
        dots = ax.scatter(x, y, c=Theta.T[0])
        r, phi = calc_coherence_and_phase(x, y)
        arrow = ax.arrow(0, 0, r*np.cos(phi), r*np.sin(phi), color='black')
        mean_dot, = ax.plot(r*np.cos(phi), r*np.sin(phi), 'o', color='black')
        if abs(K) > 0:
            title = f"coupling strength K={K:.1f}"
        else:
            title = "uncoupled"
        ax.text(-1., 1.05, title, color='gray', va='bottom')
        ax.set_aspect('equal')
        ax.set_axis_off()
        ax.set(xlim=(-1.2, 1.2), ylim=(-1.2, 1.2))
        artists.append((dots, arrow, mean_dot))


    def update(frame_number):
        updated_artists = []
        for (dots, arrow, mean_dot), K, Theta in zip(artists, K_values, theta_values):
            t = frame_delay*frame_number
            ts = timestep(t, defaultclock.dt)
            x, y = to_x_y(Theta.T[ts])
            dots.set_offsets(np.vstack([x, y]).T)
            r, phi = calc_coherence_and_phase(x, y)
            arrow.set_data(dx=r*np.cos(phi), dy=r*np.sin(phi))
            mean_dot.set_data(r*np.cos(phi), r*np.sin(phi))
            updated_artists.extend([dots, arrow, mean_dot])
        return updated_artists

    ani = animation.FuncAnimation(fig, update, frames=int(magic_network.t/frame_delay),
                                interval=20, blit=True)

    return ani

if __name__ == '__main__':
    fig, axs = plt.subplots(2, 2)
    # Manual adjustments instead of layout='tight', to avoid jumps in saved animation
    fig.subplots_adjust(left=0.025, bottom=0.025, right=0.975,  top=0.975,
                        wspace=0, hspace=0)
    K_values = [0, 1, 2, 4]
    theta_values = []
    for K in K_values:
        print(f"Running simulation for K={K:.1f}")
        theta_values.append( run_sim(K/second))
    ani = do_animation(fig, axs.flat, K_values, theta_values)

    plt.show()



import logging
import tempfile
import uuid

import pytest
from numpy.testing import assert_array_equal

from brian2 import *
from brian2.devices.cpp_standalone.device import CPPStandaloneDevice
from brian2.tests.utils import assert_allclose
from brian2.utils.logger import catch_logs


@pytest.mark.standalone_compatible
def test_spike_monitor():
    G_without_threshold = NeuronGroup(5, "x : 1")
    G = NeuronGroup(
        3,
        """
        dv/dt = rate : 1
        rate: Hz
        """,
        threshold="v>1",
        reset="v=0",
    )
    # We don't use 100 and 1000Hz, because then the membrane potential would
    # be exactly at 1 after 10 resp. 100 timesteps. Due to floating point
    # issues this will not be exact,
    G.rate = [101, 0, 1001] * Hz

    mon = SpikeMonitor(G)

    with pytest.raises(ValueError):
        SpikeMonitor(G, order=1)  # need to specify 'when' as well
    with pytest.raises(ValueError) as ex:
        SpikeMonitor(G_without_threshold)
    assert "threshold" in str(ex)

    # Creating a SpikeMonitor for a Synapses object should not work
    S = Synapses(G, G, on_pre="v += 0")
    S.connect()
    with pytest.raises(TypeError):
        SpikeMonitor(S)

    run(10 * ms)

    spike_trains = mon.spike_trains()

    assert_allclose(mon.t[mon.i == 0], [9.9] * ms)
    assert len(mon.t[mon.i == 1]) == 0
    assert_allclose(mon.t[mon.i == 2], np.arange(10) * ms + 0.9 * ms)
    assert_allclose(mon.t_[mon.i == 0], np.array([9.9 * float(ms)]))
    assert len(mon.t_[mon.i == 1]) == 0
    assert_allclose(mon.t_[mon.i == 2], (np.arange(10) + 0.9) * float(ms))
    assert_allclose(spike_trains[0], [9.9] * ms)
    assert len(spike_trains[1]) == 0
    assert_allclose(spike_trains[2], np.arange(10) * ms + 0.9 * ms)
    assert_array_equal(mon.count, np.array([1, 0, 10]))

    i, t = mon.it
    i_, t_ = mon.it_
    assert_array_equal(i, mon.i)
    assert_array_equal(i, i_)
    assert_array_equal(t, mon.t)
    assert_array_equal(t_, mon.t_)

    with pytest.raises(KeyError):
        spike_trains[3]
    with pytest.raises(KeyError):
        spike_trains[-1]
    with pytest.raises(KeyError):
        spike_trains["string"]

    # Check that indexing into the VariableView works (this fails if we do not
    # update the N variable correctly)
    assert_allclose(mon.t[:5], [0.9, 1.9, 2.9, 3.9, 4.9] * ms)


def test_spike_monitor_indexing():
    generator = SpikeGeneratorGroup(3, [0, 0, 0, 1], [0, 1, 2, 1] * ms)
    mon = SpikeMonitor(generator)
    run(3 * ms)

    assert_array_equal(mon.t["i == 1"], [1] * ms)
    assert_array_equal(mon.t[mon.i == 1], [1] * ms)
    assert_array_equal(mon.i[mon.t > 1.5 * ms], [0] * ms)
    assert_array_equal(mon.i["t > 1.5 * ms"], [0] * ms)


@pytest.mark.standalone_compatible
def test_spike_monitor_variables():
    G = NeuronGroup(
        3,
        """
        dv/dt = rate : 1
       rate : Hz
       prev_spikes : integer
       """,
        threshold="v>1",
        reset="v=0; prev_spikes += 1",
    )
    # We don't use 100 and 1000Hz, because then the membrane potential would
    # be exactly at 1 after 10 resp. 100 timesteps. Due to floating point
    # issues this will not be exact,
    G.rate = [101, 0, 1001] * Hz
    mon1 = SpikeMonitor(G, variables="prev_spikes")
    mon2 = SpikeMonitor(G, variables="prev_spikes", when="after_resets")
    run(10 * ms)
    all_values = mon1.all_values()
    prev_spikes_values = mon1.values("prev_spikes")
    assert_array_equal(mon1.prev_spikes[mon1.i == 0], [0])
    assert_array_equal(prev_spikes_values[0], [0])
    assert_array_equal(all_values["prev_spikes"][0], [0])
    assert_array_equal(mon1.prev_spikes[mon1.i == 1], [])
    assert_array_equal(prev_spikes_values[1], [])
    assert_array_equal(all_values["prev_spikes"][1], [])
    assert_array_equal(mon1.prev_spikes[mon1.i == 2], np.arange(10))
    assert_array_equal(prev_spikes_values[2], np.arange(10))
    assert_array_equal(all_values["prev_spikes"][2], np.arange(10))
    assert_array_equal(mon2.prev_spikes[mon2.i == 0], [1])
    assert_array_equal(mon2.prev_spikes[mon2.i == 1], [])
    assert_array_equal(mon2.prev_spikes[mon2.i == 2], np.arange(10) + 1)


@pytest.mark.standalone_compatible
def test_spike_monitor_get_states():
    G = NeuronGroup(
        3,
        """dv/dt = rate : 1
                          rate : Hz
                          prev_spikes : integer""",
        threshold="v>1",
        reset="v=0; prev_spikes += 1",
    )
    # We don't use 100 and 1000Hz, because then the membrane potential would
    # be exactly at 1 after 10 resp. 100 timesteps. Due to floating point
    # issues this will not be exact,
    G.rate = [101, 0, 1001] * Hz
    mon1 = SpikeMonitor(G, variables="prev_spikes")
    run(10 * ms)
    all_states = mon1.get_states()
    assert set(all_states.keys()) == {"count", "i", "t", "prev_spikes", "N"}
    assert_array_equal(all_states["count"], mon1.count[:])
    assert_array_equal(all_states["i"], mon1.i[:])
    assert_array_equal(all_states["t"], mon1.t[:])
    assert_array_equal(all_states["prev_spikes"], mon1.prev_spikes[:])
    assert_array_equal(all_states["N"], mon1.N)


@pytest.mark.standalone_compatible
def test_spike_monitor_subgroups():
    G = NeuronGroup(6, """do_spike : boolean""", threshold="do_spike")
    G.do_spike = [True, False, False, False, True, True]
    spikes_all = SpikeMonitor(G)
    spikes_1 = SpikeMonitor(G[:2])
    spikes_2 = SpikeMonitor(G[2:4])
    spikes_3 = SpikeMonitor(G[4:])
    run(defaultclock.dt)
    assert_allclose(spikes_all.i, [0, 4, 5])
    assert_allclose(spikes_all.t, [0, 0, 0] * ms)
    assert_allclose(spikes_1.i, [0])
    assert_allclose(spikes_1.t, [0] * ms)
    assert len(spikes_2.i) == 0
    assert len(spikes_2.t) == 0
    assert_allclose(spikes_3.i, [0, 1])  # recorded spike indices are relative
    assert_allclose(spikes_3.t, [0, 0] * ms)


def test_spike_monitor_bug_824():
    # See github issue #824
    if prefs.codegen.target != "numpy":
        pytest.skip("numpy-only test")

    G = NeuronGroup(6, """do_spike : boolean""", threshold="do_spike")
    G.do_spike = [True, False, False, True, False, False]
    spikes_1 = SpikeMonitor(G[:3])
    spikes_2 = SpikeMonitor(G[3:])
    run(4 * defaultclock.dt)
    assert_array_equal(spikes_1.count, [4, 0, 0])
    assert_array_equal(spikes_2.count, [4, 0, 0])


@pytest.mark.standalone_compatible
def test_event_monitor():
    G = NeuronGroup(
        3,
        """
        dv/dt = rate : 1
        rate: Hz
        """,
        events={"my_event": "v>1"},
    )
    G.run_on_event("my_event", "v=0")
    # We don't use 100 and 1000Hz, because then the membrane potential would
    # be exactly at 1 after 10 resp. 100 timesteps. Due to floating point
    # issues this will not be exact,
    G.rate = [101, 0, 1001] * Hz

    mon = EventMonitor(G, "my_event")
    net = Network(G, mon)
    net.run(10 * ms)

    event_trains = mon.event_trains()

    assert_allclose(mon.t[mon.i == 0], [9.9] * ms)
    assert len(mon.t[mon.i == 1]) == 0
    assert_allclose(mon.t[mon.i == 2], np.arange(10) * ms + 0.9 * ms)
    assert_allclose(mon.t_[mon.i == 0], np.array([9.9 * float(ms)]))
    assert len(mon.t_[mon.i == 1]) == 0
    assert_allclose(mon.t_[mon.i == 2], (np.arange(10) + 0.9) * float(ms))
    assert_allclose(event_trains[0], [9.9] * ms)
    assert len(event_trains[1]) == 0
    assert_allclose(event_trains[2], np.arange(10) * ms + 0.9 * ms)
    assert_array_equal(mon.count, np.array([1, 0, 10]))

    i, t = mon.it
    i_, t_ = mon.it_
    assert_array_equal(i, mon.i)
    assert_array_equal(i, i_)
    assert_array_equal(t, mon.t)
    assert_array_equal(t_, mon.t_)

    with pytest.raises(KeyError):
        event_trains[3]
    with pytest.raises(KeyError):
        event_trains[-1]
    with pytest.raises(KeyError):
        event_trains["string"]


@pytest.mark.standalone_compatible
def test_event_monitor_no_record():
    # Check that you can switch off recording spike times/indices
    G = NeuronGroup(
        3,
        """
        dv/dt = rate : 1
        rate: Hz
        """,
        events={"my_event": "v>1"},
        threshold="v>1",
        reset="v=0",
    )
    # We don't use 100 and 1000Hz, because then the membrane potential would
    # be exactly at 1 after 10 resp. 100 timesteps. Due to floating point
    # issues this will not be exact,
    G.rate = [101, 0, 1001] * Hz

    event_mon = EventMonitor(G, "my_event", record=False)
    event_mon2 = EventMonitor(G, "my_event", variables="rate", record=False)
    spike_mon = SpikeMonitor(G, record=False)
    spike_mon2 = SpikeMonitor(G, variables="rate", record=False)
    net = Network(G, event_mon, event_mon2, spike_mon, spike_mon2)
    net.run(10 * ms)

    # i and t should not be there
    assert "i" not in event_mon.variables
    assert "t" not in event_mon.variables
    assert "i" not in spike_mon.variables
    assert "t" not in spike_mon.variables

    assert_array_equal(event_mon.count, np.array([1, 0, 10]))
    assert_array_equal(spike_mon.count, np.array([1, 0, 10]))
    assert spike_mon.num_spikes == sum(spike_mon.count)
    assert event_mon.num_events == sum(event_mon.count)

    # Other variables should still have been recorded
    assert len(spike_mon2.rate) == spike_mon.num_spikes
    assert len(event_mon2.rate) == event_mon.num_events


@pytest.mark.standalone_compatible
def test_spike_trains():
    # An arbitrary combination of indices that has been shown to sort in an
    # unstable way with quicksort and therefore lead to out-of-order values
    # in the dictionary returned by spike_trains() of several neurons (see #725)
    generator = SpikeGeneratorGroup(
        10,
        [6, 1, 2, 4, 6, 9, 1, 4, 7, 8, 0, 6, 3, 6, 4, 8, 9, 2, 5, 3],
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19] * ms,
        dt=1 * ms,
    )
    monitor = SpikeMonitor(generator)
    run(19.1 * ms)
    trains = monitor.spike_trains()
    for idx, spike_times in trains.items():
        assert all(
            np.diff(spike_times) > 0 * ms
        ), f"Spike times for neuron {int(idx)} are not sorted"


def test_synapses_state_monitor():
    G = NeuronGroup(2, "")
    S = Synapses(G, G, "w: siemens")
    S.connect(True)
    S.w = "j*nS"

    # record from a Synapses object (all synapses connecting to neuron 1)
    synapse_mon = StateMonitor(S, "w", record=S[:, 1])
    synapse_mon2 = StateMonitor(S, "w", record=S["j==1"])

    net = Network(G, S, synapse_mon, synapse_mon2)
    net.run(10 * ms)
    # Synaptic variables
    assert_allclose(synapse_mon[S[0, 1]].w, 1 * nS)
    assert_allclose(synapse_mon.w[1], 1 * nS)
    assert_allclose(synapse_mon2[S[0, 1]].w, 1 * nS)
    assert_allclose(synapse_mon2[S["i==0 and j==1"]].w, 1 * nS)
    assert_allclose(synapse_mon2.w[1], 1 * nS)


@pytest.mark.standalone_compatible
def test_state_monitor():
    # Unique name to get the warning even for repeated runs of the test
    unique_name = f"neurongroup_{str(uuid.uuid4()).replace('-', '_')}"
    # Check that all kinds of variables can be recorded
    G = NeuronGroup(
        2,
        """
        dv/dt = -v / (10*ms) : 1
        f = clip(v, 0.1, 0.9) : 1
        rate: Hz
        """,
        threshold="v>1",
        reset="v=0",
        refractory=2 * ms,
        name=unique_name,
    )
    G.rate = [100, 1000] * Hz
    G.v = 1

    S = Synapses(G, G, "w: siemens")
    S.connect(True)
    S.w = "j*nS"

    # A bit peculiar, but in principle one should be allowed to record
    # nothing except for the time
    nothing_mon = StateMonitor(G, [], record=True)
    no_record = StateMonitor(G, "v", record=False)

    # Use a single StateMonitor
    v_mon = StateMonitor(G, "v", record=True)
    v_mon1 = StateMonitor(G, "v", record=[1])

    # Use a StateMonitor for specified variables
    multi_mon = StateMonitor(G, ["v", "f", "rate"], record=True)
    multi_mon1 = StateMonitor(G, ["v", "f", "rate"], record=[1])

    # Use a StateMonitor recording everything
    all_mon = StateMonitor(G, True, record=True)

    # Record synapses with explicit indices (the only way allowed in standalone)
    synapse_mon = StateMonitor(S, "w", record=np.arange(len(G) ** 2))

    run(10 * ms)

    # Check time recordings
    assert_array_equal(nothing_mon.t, v_mon.t)
    assert_array_equal(nothing_mon.t_, np.asarray(nothing_mon.t))
    assert_array_equal(nothing_mon.t_, v_mon.t_)
    assert_allclose(nothing_mon.t, np.arange(len(nothing_mon.t)) * defaultclock.dt)
    assert_array_equal(no_record.t, v_mon.t)

    # Check v recording
    assert_allclose(v_mon.v.T, np.exp(np.tile(-v_mon.t, (2, 1)).T / (10 * ms)))
    assert_allclose(v_mon.v_.T, np.exp(np.tile(-v_mon.t_, (2, 1)).T / float(10 * ms)))
    assert_array_equal(v_mon.v, multi_mon.v)
    assert_array_equal(v_mon.v_, multi_mon.v_)
    assert_array_equal(v_mon.v, all_mon.v)
    assert_array_equal(v_mon.v[1:2], v_mon1.v)
    assert_array_equal(multi_mon.v[1:2], multi_mon1.v)
    assert len(no_record.v) == 0

    # Other variables
    assert_array_equal(
        multi_mon.rate_.T, np.tile(np.atleast_2d(G.rate_), (multi_mon.rate.shape[1], 1))
    )
    assert_array_equal(multi_mon.rate[1:2], multi_mon1.rate)
    assert_allclose(np.clip(multi_mon.v, 0.1, 0.9), multi_mon.f)
    assert_allclose(np.clip(multi_mon1.v, 0.1, 0.9), multi_mon1.f)

    assert all(
        all_mon[0].not_refractory[:] == True
    ), "not_refractory should be True, but got(not_refractory, v):%s " % str(
        (all_mon.not_refractory, all_mon.v)
    )

    # Synapses
    assert_allclose(
        synapse_mon.w[:], np.tile(S.j[:] * nS, (synapse_mon.w[:].shape[1], 1)).T
    )


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_state_monitor_record_single_timestep():
    G = NeuronGroup(1, "dv/dt = -v/(5*ms) : 1")
    G.v = 1
    mon = StateMonitor(G, "v", record=True)
    # Recording before a run should not work
    with pytest.raises(TypeError):
        mon.record_single_timestep()
    run(0.5 * ms)
    mon.record_single_timestep()
    device.build(direct_call=False, **device.build_options)
    assert mon.t[0] == 0 * ms
    assert mon[0].v[0] == 1
    assert_allclose(mon.t[-1], 0.5 * ms)
    assert len(mon.t) == 6
    assert mon[0].v[-1] == G.v


@pytest.mark.standalone_compatible
def test_state_monitor_bigger_dt():
    # Check that it is possible to record with a slower clock, i.e. bigger dt
    G = NeuronGroup(1, "dv/dt = -v/(5*ms) : 1", method="exact")
    G.v = 1
    mon = StateMonitor(G, "v", record=True)
    slow_mon = StateMonitor(G, "v", record=True, dt=defaultclock.dt * 5)
    run(defaultclock.dt * 11)
    assert len(mon.t) == len(mon.v[0]) == 11
    assert len(slow_mon.t) == len(slow_mon.v[0]) == 3
    for timepoint in [0, 5, 10]:
        assert mon.t[timepoint] == slow_mon.t[timepoint // 5]
        assert mon.v[0, timepoint] == slow_mon.v[0, timepoint // 5]


@pytest.mark.standalone_compatible
def test_state_monitor_indexing():
    # Check indexing semantics
    G = NeuronGroup(10, "v:volt")
    G.v = np.arange(10) * volt
    mon = StateMonitor(G, "v", record=[5, 6, 7])

    run(2 * defaultclock.dt)

    assert_array_equal(mon.v, np.array([[5, 5], [6, 6], [7, 7]]) * volt)
    assert_array_equal(mon.v_, np.array([[5, 5], [6, 6], [7, 7]]))
    assert_array_equal(mon[5].v, mon.v[0])
    assert_array_equal(mon[7].v, mon.v[2])
    assert_array_equal(mon[[5, 7]].v, mon.v[[0, 2]])
    assert_array_equal(mon[np.array([5, 7])].v, mon.v[[0, 2]])

    assert_allclose(mon.t[1:], Quantity([defaultclock.dt]))

    with pytest.raises(IndexError):
        mon[8]
    with pytest.raises(TypeError):
        mon["string"]
    with pytest.raises(TypeError):
        mon[5.0]
    with pytest.raises(TypeError):
        mon[[5.0, 6.0]]


@pytest.mark.standalone_compatible
def test_state_monitor_get_states():
    G = NeuronGroup(
        2,
        """
        dv/dt = -v / (10*ms) : 1
        f = clip(v, 0.1, 0.9) : 1
        rate: Hz
        """,
        threshold="v>1",
        reset="v=0",
        refractory=2 * ms,
    )
    G.rate = [100, 1000] * Hz
    G.v = 1
    mon = StateMonitor(G, ["v", "f", "rate"], record=True)
    run(10 * defaultclock.dt)
    all_states = mon.get_states()
    assert set(all_states.keys()) == {"v", "f", "rate", "t", "N"}
    assert_array_equal(all_states["v"].T, mon.v[:])
    assert_array_equal(all_states["f"].T, mon.f[:])
    assert_array_equal(all_states["rate"].T, mon.rate[:])
    assert_array_equal(all_states["t"], mon.t[:])
    assert_array_equal(all_states["N"], mon.N)


@pytest.mark.standalone_compatible
def test_state_monitor_resize():
    # Test for issue #518 (cython did not resize the Variable object)
    G = NeuronGroup(2, "v : 1")
    mon = StateMonitor(G, "v", record=True)
    defaultclock.dt = 0.1 * ms
    run(1 * ms)
    # On standalone, the size information of the variables is only updated
    # after the variable has been accessed, so we can not only check the size
    # information of the variables object
    assert len(mon.t) == 10
    assert mon.v.shape == (2, 10)
    assert mon.variables["t"].size == 10
    # Note that the internally stored variable has the transposed shape of the
    # variable that is visible to the user
    assert mon.variables["v"].size == (10, 2)


@pytest.mark.standalone_compatible
def test_state_monitor_synapses():
    # Check that recording from synapses works correctly
    G = NeuronGroup(5, "v : 1", threshold="False", reset="v = 0")
    S1 = Synapses(G, G, "w : 1", on_pre="v_post += w")
    S1.run_regularly("w += 1")
    S1.connect(i=[0, 1], j=[2, 3])
    S1.w = "i"
    # We can check the record argument even in standalone mode, since we created
    # the synapses with an array of indices of known length
    with catch_logs() as l:
        S1_mon = StateMonitor(S1, "w", record=[0, 1])
    assert len(l) == 0

    with pytest.raises(IndexError):
        StateMonitor(S1, "w", record=[0, 2])

    S2 = Synapses(G, G, "w : 1", on_pre="v_post += w")
    S2.run_regularly("w += 1")
    S2.connect("i==j")  # Not yet executed for standalone mode
    S2.w = "i"
    with catch_logs() as l:
        S2_mon = StateMonitor(S2, "w", record=[0, 4])

    if isinstance(get_device(), CPPStandaloneDevice):
        assert len(l) == 1
        assert l[0][0] == "WARNING"
        assert l[0][1].endswith(".cannot_check_statemonitor_indices")
    else:
        assert len(l) == 0
    run(2 * defaultclock.dt)
    assert_array_equal(S1_mon.w[:], [[0, 1], [1, 2]])
    assert_array_equal(S2_mon.w[:], [[0, 1], [4, 5]])


@pytest.mark.codegen_independent
def test_statemonitor_namespace():
    # Make sure that StateMonitor is correctly inheriting its source's namespace
    G = NeuronGroup(2, "x = i + y : integer", namespace={"y": 3})
    mon = StateMonitor(G, "x", record=True)
    run(defaultclock.dt, namespace={})
    assert_array_equal(mon.x, [[3], [4]])


@pytest.mark.standalone_compatible
def test_rate_monitor_1():
    G = NeuronGroup(5, "v : 1", threshold="v>1")  # no reset
    G.v = 1.1  # All neurons spike every time step
    rate_mon = PopulationRateMonitor(G)
    run(10 * defaultclock.dt)

    assert_allclose(rate_mon.t, np.arange(10) * defaultclock.dt)
    assert_allclose(rate_mon.t_, np.arange(10) * defaultclock.dt_)
    assert_allclose(rate_mon.t, np.arange(10) * defaultclock.dt)
    assert_allclose(rate_mon.rate, np.ones(10) / defaultclock.dt)
    assert_allclose(rate_mon.rate_, np.asarray(np.ones(10) / defaultclock.dt_))
    # Check that indexing into the VariableView works (this fails if we do not
    # update the N variable correctly)
    assert_allclose(rate_mon.t[:5], np.arange(5) * defaultclock.dt)


@pytest.mark.standalone_compatible
def test_rate_monitor_2():
    G = NeuronGroup(10, "v : 1", threshold="v>1")  # no reset
    G.v["i<5"] = 1.1  # Half of the neurons fire every time step
    rate_mon = PopulationRateMonitor(G)
    net = Network(G, rate_mon)
    net.run(10 * defaultclock.dt)
    assert_allclose(rate_mon.rate, 0.5 * np.ones(10) / defaultclock.dt)
    assert_allclose(rate_mon.rate_, 0.5 * np.asarray(np.ones(10) / defaultclock.dt_))


@pytest.mark.codegen_independent
def test_rate_monitor_smoothed_rate():
    # Test the filter response by having a single spiking neuron
    G = SpikeGeneratorGroup(1, [0], [1] * ms)
    r_mon = PopulationRateMonitor(G)
    run(3 * ms)
    index = int(np.round(1 * ms / defaultclock.dt))
    except_index = np.array([idx for idx in range(len(r_mon.rate)) if idx != index])
    assert_array_equal(r_mon.rate[except_index], 0 * Hz)
    assert_allclose(r_mon.rate[index], 1 / defaultclock.dt)
    ### Flat window
    # Using a flat window of size = dt should not change anything
    assert_allclose(r_mon.rate, r_mon.smooth_rate(window="flat", width=defaultclock.dt))
    smoothed = r_mon.smooth_rate(window="flat", width=5 * defaultclock.dt)
    assert_array_equal(smoothed[: index - 2], 0 * Hz)
    assert_array_equal(smoothed[index + 3 :], 0 * Hz)
    assert_allclose(smoothed[index - 2 : index + 3], 0.2 / defaultclock.dt)
    with catch_logs(log_level=logging.INFO):
        smoothed2 = r_mon.smooth_rate(window="flat", width=5.4 * defaultclock.dt)
        assert_array_equal(smoothed, smoothed2)

    ### Gaussian window
    width = 5 * defaultclock.dt
    smoothed = r_mon.smooth_rate(window="gaussian", width=width)
    # 0 outside of window
    assert_array_equal(smoothed[: index - 10], 0 * Hz)
    assert_array_equal(smoothed[index + 11 :], 0 * Hz)
    # Gaussian around the spike
    gaussian = np.exp(
        -((r_mon.t[index - 10 : index + 11] - 1 * ms) ** 2) / (2 * width**2)
    )
    gaussian /= sum(gaussian)
    assert_allclose(smoothed[index - 10 : index + 11], 1 / defaultclock.dt * gaussian)

    ### Arbitrary window
    window = np.ones(5)
    smoothed_flat = r_mon.smooth_rate(window="flat", width=5 * defaultclock.dt)
    smoothed_custom = r_mon.smooth_rate(window=window)
    assert_allclose(smoothed_flat, smoothed_custom)


@pytest.mark.codegen_independent
def test_rate_monitor_smoothed_rate_incorrect():
    # Test the filter response by having a single spiking neuron
    G = SpikeGeneratorGroup(1, [0], [1] * ms)
    r_mon = PopulationRateMonitor(G)
    run(2 * ms)

    with pytest.raises(TypeError):
        r_mon.smooth_rate(window="flat")  # no width
    with pytest.raises(TypeError):
        r_mon.smooth_rate(window=np.ones(5), width=1 * ms)
    with pytest.raises(NotImplementedError):
        r_mon.smooth_rate(window="unknown", width=1 * ms)
    with pytest.raises(TypeError):
        r_mon.smooth_rate(window=object())
    with pytest.raises(TypeError):
        r_mon.smooth_rate(window=np.ones(5, 2))
    with pytest.raises(TypeError):
        r_mon.smooth_rate(window=np.ones(4))  # even number


@pytest.mark.standalone_compatible
def test_rate_monitor_get_states():
    G = NeuronGroup(5, "v : 1", threshold="v>1")  # no reset
    G.v = 1.1  # All neurons spike every time step
    rate_mon = PopulationRateMonitor(G)
    run(10 * defaultclock.dt)
    all_states = rate_mon.get_states()
    assert set(all_states.keys()) == {"rate", "t", "N"}
    assert_array_equal(all_states["rate"], rate_mon.rate[:])
    assert_array_equal(all_states["t"], rate_mon.t[:])
    assert_array_equal(all_states["N"], rate_mon.N)


@pytest.mark.standalone_compatible
def test_rate_monitor_subgroups():
    old_dt = defaultclock.dt
    defaultclock.dt = 0.01 * ms
    G = NeuronGroup(
        4,
        """
        dv/dt = rate : 1
        rate : Hz
        """,
        threshold="v>0.999",
        reset="v=0",
    )
    G.rate = [100, 200, 400, 800] * Hz
    rate_all = PopulationRateMonitor(G)
    rate_1 = PopulationRateMonitor(G[:2])
    rate_2 = PopulationRateMonitor(G[2:])
    run(1 * second)
    assert_allclose(mean(G.rate_[:]), mean(rate_all.rate_[:]))
    assert_allclose(mean(G.rate_[:2]), mean(rate_1.rate_[:]))
    assert_allclose(mean(G.rate_[2:]), mean(rate_2.rate_[:]))

    defaultclock.dt = old_dt


@pytest.mark.standalone_compatible
def test_rate_monitor_subgroups_2():
    G = NeuronGroup(6, """do_spike : boolean""", threshold="do_spike")
    G.do_spike = [True, False, False, False, True, True]
    rate_all = PopulationRateMonitor(G)
    rate_1 = PopulationRateMonitor(G[:2])
    rate_2 = PopulationRateMonitor(G[2:4])
    rate_3 = PopulationRateMonitor(G[4:])
    run(2 * defaultclock.dt)
    assert_allclose(rate_all.rate, 0.5 / defaultclock.dt)
    assert_allclose(rate_1.rate, 0.5 / defaultclock.dt)
    assert_allclose(rate_2.rate, 0 * Hz)
    assert_allclose(rate_3.rate, 1 / defaultclock.dt)


@pytest.mark.codegen_independent
def test_monitor_str_repr():
    # Basic test that string representations are not empty
    G = NeuronGroup(2, "dv/dt = -v/(10*ms) : 1", threshold="v>1", reset="v=0")
    spike_mon = SpikeMonitor(G)
    state_mon = StateMonitor(G, "v", record=True)
    rate_mon = PopulationRateMonitor(G)
    for obj in [spike_mon, state_mon, rate_mon]:
        assert len(str(obj))
        assert len(repr(obj))


if __name__ == "__main__":
    from _pytest.outcomes import Skipped

    test_spike_monitor()
    test_spike_monitor_indexing()
    test_spike_monitor_get_states()
    test_spike_monitor_subgroups()
    try:
        test_spike_monitor_bug_824()
    except Skipped:
        pass
    test_spike_monitor_variables()
    test_event_monitor()
    test_event_monitor_no_record()
    test_spike_trains()
    test_synapses_state_monitor()
    test_state_monitor()
    test_state_monitor_record_single_timestep()
    test_state_monitor_get_states()
    test_state_monitor_indexing()
    test_state_monitor_resize()
    test_rate_monitor_1()
    test_rate_monitor_2()
    test_rate_monitor_smoothed_rate()
    test_rate_monitor_smoothed_rate_incorrect()
    test_rate_monitor_get_states()
    test_rate_monitor_subgroups()
    test_rate_monitor_subgroups_2()
    test_monitor_str_repr()
